from flask import Flask

from . import rpc
from . import script


def create_app() -> Flask:
    """
        Creates a new Flask instance for the objection API

        :return:
    """

    app = Flask(__name__)
    app.register_blueprint(rpc.bp)
    app.register_blueprint(script.bp)

    return app



import click

from objection.state.connection import state_connection


def string_canary(args: list) -> None:
    """
        Monitors for a string canary argument and reports when
        it is found.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: android monitor canary <value> (optional: <filter>)', bold=True)
        return

    target_class = args[0]

    api = state_connection.get_api()
    api.android_live_print_class_instances(target_class)



from objection.state.connection import state_connection


def _should_be_quiet(args: list) -> bool:
    """
        Checks if --quiet is part of the
        commands arguments.

        :param args:
        :return:
    """

    return '--quiet' in args


def ios_disable(args: list = None) -> None:
    """
        Starts a new objection job that hooks common classes and functions,
        applying new logic in an attempt to bypass SSL pinning.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.ios_pinning_disable(_should_be_quiet(args))



import os
from abc import ABC

from objection.state.connection import state_connection
from objection.utils.helpers import debug_print
from ..state.api import api_state


class Plugin(ABC):
    """ Plugin object to extend for development of custom functionality """

    def __init__(self, plugin_file: str, namespace: str, implementation: dict):
        """
            Start a new plugin instance.

            :param plugin_file:
            :param namespace:
            :param implementation:
        """

        self.namespace = namespace
        self.implementation = implementation
        self.plugin_file = plugin_file

        # plugin properties
        if not hasattr(self, 'script_src'):
            self.script_src = None
        if not hasattr(self, 'script_path'):
            self.script_path = None
        if not hasattr(self, 'on_message_handler'):
            self.on_message_handler = None

        self.agent = None
        self.session = None
        self.script = None
        self.api = None

        self._prepare_source()
        self._append_to_api()

    def _prepare_source(self):
        """
            Prepares the self.script_src attribute based on a few rules.

            If the script source is already set, simply return as there is
                nothing for us to do.
            If the script path is set, read that and populate the script_src
                attribute.
            If neither script_src not script_path is set, attempt to read the
                index.js that lives next to the plugin file.

            If all of the above fail, simply return, writing a debug warning
                no script source could be found.

            :return:
        """

        if self.script_src:
            return

        if self.script_path:
            self.script_path = os.path.abspath(self.script_path)
            with open(self.script_path, 'r', encoding='utf-8') as f:
                self.script_src = '\n'.join(f.readlines())
            return

        possible_src = os.path.abspath(os.path.join(
            os.path.abspath(os.path.dirname(self.plugin_file)), 'index.js'))
        if os.path.exists(possible_src):
            self.script_path = possible_src
            with open(self.script_path, 'r', encoding='utf-8') as f:
                self.script_src = '\n'.join(f.readlines())
            return

        debug_print('[warning] No Fridascript could be found for plugin {0}'.format(self.namespace))

    def inject(self) -> None:
        """
            Injects the script sources in a new Frida session.

            :return:
        """

        if not self.script_src:
            raise Exception('Unable to discover Frida script source to inject')

        if not self.agent:
            self.agent = state_connection.get_agent()

        self.session = self.agent.device.attach(self.agent.pid)
        self.script = self.session.create_script(source=self.script_src)

        # check for a custom message handler, otherwise fallback
        # to the default objection handler
        self.script.on('message',
                       self.on_message_handler if self.on_message_handler else self.agent.handlers.script_on_message)

        self.script.load()
        self.api = self.script.exports

    def _append_to_api(self):
        """
            If the http_api() function is defined in the child class, take
            it's return (it should always return a flask.Blueprint) and append
            it to the existing blueprints in objections core API.

            The ApiState class will handle the loading and starting of the API
            with them included.

            :return:
        """

        if not hasattr(self, 'http_api'):
            return

        if not callable(getattr(self, 'http_api')):
            raise Exception('The http_api property must be a function returning a Flask Blueprint')

        api_state.append_api_blueprint(getattr(self, 'http_api')())



import click

from objection.state.connection import state_connection
from ..state.device import device_state, Ios, Android


def alert(args: list = None) -> None:
    """
        Displays an alert message via a popup or a Toast message
        on the mobile device.

        :param args:
        :return:
    """

    if len(args) <= 0:
        message = 'objection!'
    else:
        message = args[0]

    if device_state.platform == Ios:
        _alert_ios(message)

    if device_state.platform == Android:
        pass


def _alert_ios(message: str):
    """
        Display an alert on iOS

        :param message:
        :return:
    """

    api = state_connection.get_api()
    api.ios_ui_alert(message)


def ios_screenshot(args: list = None) -> None:
    """
        Take an iOS screenshot.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: ios ui screenshot <local png destination>', bold=True)
        return

    destination = args[0]

    if not destination.endswith('.png'):
        destination = destination + '.png'

    api = state_connection.get_api()
    png = api.ios_ui_screenshot()

    with open(destination, 'wb') as f:
        f.write(png)

    click.secho('Screenshot saved to: {0}'.format(destination), fg='green')


def dump_ios_ui(args: list = None) -> None:
    """
        Dumps the current iOS user interface in a serialized form.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    ui = api.ios_ui_window_dump()

    click.secho(ui)


def bypass_touchid(args: list = None) -> None:
    """
        Starts a new objection job that hooks into the iOS TouchID
        classes, replacing the verification logic to always pass.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.ios_ui_biometrics_bypass()


def android_screenshot(args: list = None) -> None:
    """
        Take an Android screenshot.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: android ui screenshot <local png destination>', bold=True)
        return

    # add the .png extension if it does not already exist
    destination = args[0] if args[0].endswith('.png') else args[0] + '.png'

    # download the file
    api = state_connection.get_api()
    data = api.android_ui_screenshot()

    image = bytearray(map(lambda x: x % 256, data))

    with open(destination, 'wb') as f:
        f.write(image)

    click.secho('Screenshot saved to: {0}'.format(destination), fg='green')


def android_flag_secure(args: list = None) -> None:
    """
        Control FLAG_SECURE of the current Activity, allowing or disallowing
        the use of hardware key combinations and screencap to take screenshots.

        :param args:
        :return:
    """

    if len(args) <= 0 or args[0] not in ('true', 'false'):
        click.secho('Usage: android ui FLAG_SECURE <true/false>', bold=True)
        return

    api = state_connection.get_api()
    api.android_ui_set_flag_secure(args[0])



import os

import click

from ..state.app import app_state


def history(args: list) -> None:
    """
        Lists the commands that have been run in the current session.

        :param args:
        :return:
    """

    click.secho('Unique commands run in current session:', dim=True)

    for command in app_state.successful_commands:
        click.secho(command)


def save(args: list) -> None:
    """
        Save the current sessions command history to a file.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: commands save <local destination>', bold=True)
        return

    destination = os.path.expanduser(args[0]) if args[0].startswith('~') else args[0]

    with open(destination, 'w') as f:
        for command in app_state.successful_commands:
            f.write('{0}\n'.format(command))

    click.secho('Saved commands to: {0}'.format(destination), fg='green')


def clear(args: list) -> None:
    """
        Clears the current sessions command history.

        :param args:
        :return:
    """

    app_state.clear_command_history()
    click.secho('Command history cleared.', fg='green')



import datetime
import lzma
import os
import plistlib
import shutil
import tempfile
import zipfile

import click
import delegator
import requests

from .base import BasePlatformGadget, BasePlatformPatcher, objection_path
from .github import Github


class IosGadget(BasePlatformGadget):
    """ Class used to work with the iOS Frida Gadget """

    ios_dylib_path = os.path.join(objection_path, 'ios')
    ios_dylib_gadget_path = os.path.join(ios_dylib_path, 'FridaGadget.dylib')
    ios_dylib_gadget_archive_path = os.path.join(ios_dylib_path, 'FridaGadget.dylib.xz')

    def __init__(self, github: Github) -> None:
        """
            Build a new instance, ensuring that the paths needed
            are available.

            :param github:
        """

        super(IosGadget, self).__init__(github)

        # ensure we have the ios gadget path available
        if not os.path.exists(self.ios_dylib_path):
            os.makedirs(self.ios_dylib_path)

    def get_gadget_path(self) -> str:
        """
            Returns the path on disk where the iOS FridaGadget
            can be found.

            :return:
        """

        return self.ios_dylib_gadget_path

    def gadget_exists(self):
        """
            Checks if the iOS gadget exists on disk.

            :return:
        """

        return os.path.exists(self.ios_dylib_gadget_path)

    def download(self):
        """
            Downloads the latest iOS gadget.

            :return:
        """

        download_url = self._get_download_url()
        click.secho('Downloading from: {0}'.format(download_url), dim=True)

        # stream the download using requests
        dylib = requests.get(download_url, stream=True)

        # save the requests stream to file
        try:
            with open(self.ios_dylib_gadget_archive_path, 'wb') as f:
                click.secho('Downloading iOS dylib to {0}...'.format(self.ios_dylib_gadget_archive_path),
                            fg='green', dim=True)

                shutil.copyfileobj(dylib.raw, f)
        except Exception:
            if os.path.isfile(self.ios_dylib_gadget_archive_path):
                os.remove(self.ios_dylib_gadget_archive_path)

        return self

    def _get_download_url(self) -> str:
        """
            Determines the download URL to use for the iOS
            gadget.

            :return:
        """

        url = ''

        for asset in self.github.get_assets():
            if 'ios-universal.dylib.xz' in asset['name']:
                url = asset['browser_download_url']

        if not url:
            click.secho('Unable to determine URL to download the dylib', fg='red')
            raise Exception('Unable to determine URL for iOS gadget download.')

        return url

    def unpack(self):
        """
            Unpacks a downloaded .xz gadget.

            :return:
        """

        click.secho('Unpacking {0}...'.format(self.ios_dylib_gadget_archive_path), dim=True)

        try:
            with lzma.open(self.ios_dylib_gadget_archive_path) as f:
                with open(self.ios_dylib_gadget_path, 'wb') as g:
                    g.write(f.read())
        except Exception:
            if os.path.isfile(self.ios_dylib_gadget_archive_path):
                os.remove(self.ios_dylib_gadget_archive_path)
            if os.path.isfile(self.ios_dylib_gadget_path):
                os.remove(self.ios_dylib_gadget_path)

        return self

    def cleanup(self):
        """
            Cleans up a downloaded iOS .xz gadget.

            :return:
        """

        click.secho('Cleaning up downloaded archives...', dim=True)

        os.remove(self.ios_dylib_gadget_archive_path)


class IosPatcher(BasePlatformPatcher):
    """ Class used to Patch iOS applications """

    required_commands = {
        'xcodebuild': {
            'installation': 'Install XCode on macOS via the Appstore'
        },
        'applesign': {
            'installation': 'npm install -g applesign'
        },
        'insert_dylib': {
            'installation': ('git clone https://github.com/Tyilo/insert_dylib && cd insert_dylib &&'
                             'xcodebuild && cp build/Release/insert_dylib /usr/local/bin/insert_dylib')
        },
        'codesign': {
            'installation': 'Part of XCode'
        },
        'security': {
            'installation': 'macOS builtin command'
        },
        'zip': {
            'installation': 'macOS builtin command'
        },
        'unzip': {
            'installation': 'macOS builtin command'
        },
        'plutil': {
            'installation': 'macOS builtin command'
        },
    }

    def __init__(self, skip_cleanup: bool = False):
        """
            Init a new instance of the IosPatcher class.
        """

        super(IosPatcher, self).__init__()

        self.provision_file = None
        self.payload_directory = None
        self.app_folder = None
        self.app_binary = None
        self.patched_ipa_path = None
        self.patched_codesigned_ipa_path = None
        self.skip_cleanup = skip_cleanup
        self.bundle_id = None

        # temp_file to copy an IPA to
        _, self.temp_file = tempfile.mkstemp(suffix='.ipa')

        # a working directory to extract the IPA to
        self.temp_directory = os.path.dirname(self.temp_file)

        # cleanup the temp_directory to work with
        self._cleanup_extracted_data()

    def set_provsioning_profile(self, provision_file: str = None, bundle_id: str = None) -> None:
        """
            Sets the provision file to use during patching.

            :param bundle_id:
            :param provision_file:
            :return:
        """

        # have provision file? set it and be done
        if provision_file:
            self.provision_file = provision_file

            if bundle_id:
                click.secho('Setting bundleid to specified value: {}'.format(bundle_id), dim=True)
                self.bundle_id = bundle_id
            else:
                self._set_bundle_id_from_profile()

            return

        click.secho('No provision file specified, searching for one...', bold=True)

        # locate a valid mobile provision on disk in: ~/Library/Developer/Xcode/DerivedData/
        possible_provisions = [os.path.join(dp, f) for dp, dn, fn in
                               os.walk(os.path.expanduser('~/Library/Developer/Xcode/DerivedData/'))
                               for f in fn if 'embedded.mobileprovision' in f]

        if len(possible_provisions) <= 0:
            message = 'No provisioning files found. Please specify one or generate one by building an app.'
            click.secho(message, fg='red')
            raise Exception(message)

        # we have some provisioning profiles, lets find the one
        # with the most days left
        current_time = datetime.datetime.now()
        expirations = {}

        for pf in possible_provisions:
            _, decoded_location = tempfile.mkstemp('decoded_provision')

            # Decode the mobile provision using macOS's security cms tool
            delegator.run(self.list2cmdline([
                self.required_commands['security']['location'],
                'cms', '-D', '-i', pf,
                '-o', decoded_location
            ]), timeout=self.command_run_timeout)

            # read the expiration date from the profile
            with open(decoded_location, 'rb') as f:
                parsed_data = plistlib.load(f)

                if parsed_data['ExpirationDate'] > current_time:
                    expirations[pf] = parsed_data['ExpirationDate'] - current_time
                    click.secho('Found provision file {0} expiring in {1}'.format(pf, expirations[pf]), dim=True)

            # cleanup the temp path
            os.remove(decoded_location)

        # ensure that we got some valid mobileprovisions to work with
        if len(expirations) <= 0:
            message = 'Could not find a non-expired provisioning file. Please specify or generate one.'

            click.secho(message, fg='red')
            raise Exception(message)

        # sort the results so that the mobileprovision with the most time is at
        # the top of the list
        click.secho('Found a valid provisioning profile', fg='green', bold=True)
        self.provision_file = sorted(expirations, key=expirations.get, reverse=True)[0]

        if bundle_id:
            click.secho('Setting bundleid to specified value: {}'.format(bundle_id), dim=True)
            self.bundle_id = bundle_id
        else:
            self._set_bundle_id_from_profile()

    def extract_ipa(self, unzip_unicode, ipa_source: str) -> None:
        """
            Extracts a source IPA into the temporary directories.

            :param ipa_source:
            :param unzip_unicode:
            :return:
        """

        # copy the original ipa to the temp directory.
        shutil.copyfile(ipa_source, self.temp_file)

        if unzip_unicode:
            with zipfile.ZipFile(self.temp_file, 'r') as ipa:
                for info in ipa.infolist():
                    info.filename = info.filename.encode('cp437').decode('utf-8')
                    ipa.extract(info, self.temp_directory)
        else:
            # extract the IPA this should result in a 'Payload' directory
            ipa = zipfile.ZipFile(self.temp_file, 'r')
            ipa.extractall(self.temp_directory)
            ipa.close()

        # check what is in the Payload directory
        self.payload_directory = os.listdir(os.path.join(self.temp_directory, 'Payload'))
        if len(self.payload_directory) > 1:
            click.secho('Warning: Payload folder has more than one file, this is unexpected.', fg='yellow')

        # get the folder that ends with .app. This is where we will be patching
        # the executable with FridaGadget
        app_name = ''.join([x for x in self.payload_directory if x.endswith('.app')])
        click.secho('Working with app: {0}'.format(app_name))

        self.app_folder = os.path.join(self.temp_directory, 'Payload', app_name)

    def set_application_binary(self, binary: str = None) -> None:
        """
            Sets the binary that will be patched.

            If a binary is not defined, the applications Info.plist is parsed
            and the CFBundleIdentifier key read.

            :param binary:
            :return:
        """

        if binary is not None:
            click.secho('Using user provided binary name of: {0}'.format(binary))
            self.app_binary = os.path.join(self.app_folder, binary)

            return

        with open(os.path.join(self.app_folder, 'Info.plist'), 'rb') as f:
            info_plist = plistlib.load(f)

        # print the bundle identifier
        click.secho('Bundle identifier is: {0}'.format(info_plist['CFBundleIdentifier']),
                    fg='green', bold=True)

        self.app_binary = os.path.join(self.app_folder, info_plist['CFBundleExecutable'])

    def patch_and_codesign_binary(self, frida_gadget: str, codesign_signature: str, gadget_config: str) -> None:
        """
            Patches an iOS binary to load a Frida gadget on startup.

            Any other dylibs within the application will also be code signed with
            the same signature used for the FridaGadget itself.

            :param frida_gadget:
            :param codesign_signature:
            :param gadget_config:
            :return:
        """

        if not self.app_binary:
            raise Exception('The applications binary should be set first.')

        if not self.app_folder:
            raise Exception('The application should be extracted first.')

        # create a Frameworks directory if it does not already exist
        if not os.path.exists(os.path.join(self.app_folder, 'Frameworks')):
            click.secho('Creating Frameworks directory for FridaGadget...', fg='green')
            os.mkdir(os.path.join(self.app_folder, 'Frameworks'))

        # copy the frida gadget to the applications Frameworks directory
        shutil.copyfile(frida_gadget, os.path.join(self.app_folder, 'Frameworks', 'FridaGadget.dylib'))
        if gadget_config:
            click.secho('Copying Gadget Config to Frameworks path...', fg='green', dim=True)
            shutil.copyfile(gadget_config, os.path.join(self.app_folder, 'Frameworks', 'FridaGadget.config'))

        # patch the app binary
        load_library_output = delegator.run(self.list2cmdline([
            self.required_commands['insert_dylib']['location'],
            '--strip-codesig',
            '--inplace',
            '@executable_path/Frameworks/FridaGadget.dylib',
            self.app_binary
        ]), timeout=self.command_run_timeout)

        # check if the insert_dylib call may have failed
        if 'Added LC_LOAD_DYLIB' not in load_library_output.out:
            click.secho('Injecting the load library to {0} might have failed.'.format(self.app_binary),
                        fg='yellow')
            click.secho(load_library_output.out, fg='red', dim=True)
            click.secho(load_library_output.err, fg='red')

        # get the paths of all of the .dylib files in this applications
        # bundle. we will have to codesign all of them and not just the
        # frida gadget
        dylibs_to_sign = [
            os.path.join(dp, f) for dp, dn, fn in os.walk(self.app_folder) for f in fn if f.endswith('.dylib')]

        # codesign the dylibs in this bundle
        click.secho('Codesigning {0} .dylib\'s with signature {1}'.format(len(dylibs_to_sign), codesign_signature),
                    fg='green')
        for dylib in dylibs_to_sign:
            click.secho('Code signing: {0}'.format(os.path.basename(dylib)), dim=True)
            delegator.run(self.list2cmdline([
                self.required_commands['codesign']['location'],
                '-f',
                '-v',
                '-s',
                codesign_signature,
                dylib]))

    def archive_and_codesign(self, original_name: str, codesign_signature: str) -> None:
        """
            Creates a new archive of the patched IPA.

            :param original_name:
            :param codesign_signature:
            :return:
        """

        click.secho('Creating new archive with patched contents...', dim=True)
        self.patched_ipa_path = os.path.join(
            self.temp_directory, os.path.basename(
                '{0}-frida.ipa'.format(os.path.splitext(original_name)[0])))

        def zipdir(path, ziph):
            # ziph is a zipfile handle
            for root, dirs, files in os.walk(path):
                for fi in files:
                    ziph.write(os.path.join(root, fi),
                               os.path.relpath(os.path.join(root, fi), os.path.join(path, '..')))

        zipf = zipfile.ZipFile(self.patched_ipa_path, 'w')
        zipdir(os.path.join(self.temp_directory, 'Payload'), zipf)
        zipf.close()

        # codesign the new ipa
        click.secho('Codesigning patched IPA...', fg='green')
        self.patched_codesigned_ipa_path = os.path.join(self.temp_directory, os.path.basename(
            '{0}-frida-codesigned.ipa'.format(os.path.splitext(original_name)[0])))

        ipa_codesign = delegator.run(self.list2cmdline([
            self.required_commands['applesign']['location'],
            '--identity',
            codesign_signature,
            '--mobileprovision',
            self.provision_file,
            '--bundleid',
            self.bundle_id,
            '--clone-entitlements',
            '--output',
            self.patched_codesigned_ipa_path,
            self.patched_ipa_path
        ]), timeout=self.command_run_timeout)

        click.secho(ipa_codesign.err, dim=True)

    def get_patched_ipa_path(self) -> str:
        """
            Returns the path where the final patched IPA would be.

            :return:
        """

        return self.patched_codesigned_ipa_path

    def _set_bundle_id_from_profile(self):
        """
            Extracts and sets a bundle id from a decoded mobileprovision

            :return:
        """

        if not self.provision_file:
            click.secho('Provisioning profile not set. Skipping bundleid extraction', dim=True)
            return

        _, decoded_location = tempfile.mkstemp('decoded_provision')

        # Decode the mobile provision using macOS's security cms tool
        delegator.run(self.list2cmdline([
            self.required_commands['security']['location'],
            'cms', '-D', '-i', self.provision_file,
            '-o', decoded_location
        ]), timeout=self.command_run_timeout)

        # https://stackoverflow.com/a/66820375
        # security cms -D -i your.mobileprovision | plutil -extract
        #   Entitlements.application-identifier xml1 -o - - | grep string |
        #   sed 's/^<string>[^\.]*\.\(.*\)<\/string>$/\1/g'
        c = delegator.run(self.list2cmdline([
            'cat', decoded_location
        ]), timeout=self.command_run_timeout).pipe(self.list2cmdline([
            self.required_commands['plutil']['location'],
            '-extract', 'Entitlements.application-identifier', 'xml1', '-o', '-', '-'
        ]), timeout=self.command_run_timeout).pipe(self.list2cmdline([
            'grep', 'string'
        ]), timeout=self.command_run_timeout).pipe(self.list2cmdline([
            'sed', r's/^<string>[^\.]*\.\(.*\)<\/string>$/\1/g'
        ]), timeout=self.command_run_timeout)

        if len(c.out) > 0:
            self.bundle_id = c.out.strip()

        click.secho('Mobile provision bundle identifier is: {}'.format(self.bundle_id), dim=True)

        # cleanup the temp path
        os.remove(decoded_location)

    def _cleanup_extracted_data(self) -> None:
        """
            Small helper method to cleanup temporary files created
            when an older IPA was extracted.

            :return:
        """

        p = os.path.join(self.temp_directory, 'Payload')
        shutil.rmtree(p, ignore_errors=True)

    def __del__(self):
        """
            Cleanup after ourselves.

            :return:
        """

        if self.skip_cleanup:
            click.secho('Not cleaning up temporary files', dim=True)
            return

        click.secho('Cleaning up temp files...', dim=True)

        try:

            self._cleanup_extracted_data()
            os.remove(self.temp_file)
            os.remove(self.patched_ipa_path)
            os.remove(self.patched_codesigned_ipa_path)

        except Exception as err:
            click.secho('Failed to cleanup with error: {0}'.format(err), fg='red', dim=True)



import click

from objection.state.connection import state_connection


def get(args: list = None) -> None:
    """
        Gets all of the values stored in NSUserDefaults and prints
        them to screen.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    defaults = api.ios_nsuser_defaults_get()

    click.secho(defaults, bold=True)



import click
from tabulate import tabulate

from objection.state.connection import state_connection


def info(args: list) -> None:
    """
        Gets information about binaries and frameworks.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    binary_info = api.ios_binary_info()

    click.secho(tabulate(
        [[
            name,
            information['type'],
            information['encrypted'],
            information['pie'],
            information['arc'],
            information['canary'],
            information['stackExec'],
            information['rootSafe']
        ] for name, information in binary_info.items()],
        headers=['Name', 'Type', 'Encrypted', 'PIE', 'ARC', 'Canary', 'Stack Exec', 'RootSafe'],
    ))



import click
from tabulate import tabulate

from objection.state.connection import state_connection


def dump(args: list = None) -> None:
    """
        Dumps credentials stored in NSURLCredentialStorage

        :param args:
        :return:
    """

    api = state_connection.get_api()
    cookies = api.ios_credential_storage()

    click.secho(tabulate(
        [[
            entry['protocol'],
            entry['host'],
            entry['port'],
            entry['authMethod'].replace('NSURLAuthenticationMethod', ''),
            entry['user'],
            entry['password'],
        ] for entry in cookies], headers=[
            'Protocol', 'Host', 'Port', 'Authentication Method', 'User', 'Password'
        ],
    ))



class StateConnection(object):
    """ A class controlling the connection state of a device. """

    def __init__(self) -> None:
        """
            Init a new connection state, defaulting to a USB
            connection.
        """

        self.network = False
        self.host = None
        self.port = None
        self.device_type = 'usb'
        self.device_id = None

        self.spawn = False
        self.no_pause = False
        self.foremost = False
        self.debugger = False

        self.name = None
        self.agent = None
        self.api = None
        self.uid = None

    def use_usb(self) -> None:
        """
            Sets the values required to have a USB connection.

            :return:
        """

        self.network = False
        self.device_type = 'usb'

    def use_network(self) -> None:
        """
            Sets the values required to have a Network connection.

            :return:
        """

        self.network = True
        self.device_type = 'remote'

    def use_local(self) -> None:
        """
            Sets the values required to have a local connection.

            :return:
        """

        self.network = False
        self.device_type = 'local'

    def get_comms_type(self) -> int:
        """
            Returns the currently configured connection type.

            :return:
        """

    def get_api(self):
        """
            Return a Frida RPC API session

            :return:
        """

        if not self.agent:
            raise Exception('No session available to get API')

        return self.agent.exports()

    def set_agent(self, agent):
        """
            Sets the active agent to use for communications.

            :param agent:
            :return:
        """

        self.agent = agent

    def get_agent(self):

        if not self.agent:
            raise Exception('No Agent available')

        return self.agent

    def __repr__(self) -> str:
        return f'<State DevSerial: {self.device_id}, Network:{self.network}, Host:{self.host}, Port:{self.port}'


state_connection = StateConnection()



import json

import click
from tabulate import tabulate

from objection.state.connection import state_connection


def _should_output_json(args: list) -> bool:
    """
        Checks if --json is in the list of tokens received from the
        command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--json' in args


def _should_do_smart_decode(args: list) -> bool:
    """
        Checks if --smart is in the list of tokens received from the
        command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--smart' in args


def _data_flag_has_identifier(args: list) -> bool:
    """
        Checks that if the data flag is specified, an identifier
        is also passed.

        :param args:
        :return:
    """

    if '--data' in args:
        return any(x in args for x in ['--service', '--account'])

    return True


def _get_flag_value(args: list, flag: str) -> str:
    """
        Returns the value for a flag.

        :param args:
        :param flag:
        :return:
    """

    return args[args.index(flag) + 1] if flag in args else None


def dump(args: list = None) -> None:
    """
        Dump the iOS keychain

        :param args:
        :return:
    """

    if _should_output_json(args) and len(args) < 2:
        click.secho('Usage: ios keychain dump (--json <local destination>)', bold=True)
        return

    click.secho('Note: You may be asked to authenticate using the devices passcode or TouchID')

    if not _should_output_json(args):
        click.secho('Save the output by adding `--json keychain.json` to this command', dim=True)

    click.secho('Dumping the iOS keychain...', dim=True)
    api = state_connection.get_api()
    keychain = api.ios_keychain_list(_should_do_smart_decode(args))

    if _should_output_json(args):
        destination = args[1]

        click.secho('Writing keychain as json to {0}...'.format(destination), dim=True)

        with open(destination, 'w') as f:
            f.write(json.dumps(keychain, indent=2))

        click.secho('Dumped keychain to: {0}'.format(destination), fg='green')
        return

    # Just dump it to the screen
    click.secho(tabulate(
        [[
            entry['create_date'],
            entry['accessible_attribute'].replace('kSecAttrAccessible',
                                                  '') if 'accessible_attribute' in entry else None,
            entry['access_control'],
            entry['item_class'].replace('kSecClassGeneric', ''),
            entry['account'],
            entry['service'],
            entry['data']
        ] for entry in keychain], headers=['Created', 'Accessible', 'ACL', 'Type', 'Account', 'Service', 'Data'],
    ))


def dump_raw(args: list = None) -> None:
    """
        Dump the iOS keychain, but without any parsing.
        The agent will output the entries it finds here.

        :param args:
        :return:
    """

    click.secho('Note: You may be asked to authenticate using the devices passcode or TouchID')
    click.secho('Dumping the iOS keychain...', dim=True)
    api = state_connection.get_api()
    api.ios_keychain_list_raw()


def clear(args: list = None) -> None:
    """
        Clear the iOS keychain.

        :param args:
        :return:
    """

    if not click.confirm('Are you sure you want to clear the iOS keychain?'):
        return

    click.secho('Clearing the keychain...', dim=True)

    api = state_connection.get_api()
    api.ios_keychain_empty()

    click.secho('Keychain cleared', fg='green')


def remove(args: list) -> None:
    """
        Remove matching keychain entries from the keychain

        :param args:
        :return:
    """

    account = _get_flag_value(args, '--account')
    service = _get_flag_value(args, '--service')

    if not account or not service:
        click.secho('Either --account or --service is not set. We need both', bold=True)
        return

    click.secho('Removing entry from the iOS keychain...', dim=True)
    click.secho('Account:  {0}'.format(account), dim=True)
    click.secho('Service:  {0}'.format(service), dim=True)

    api = state_connection.get_api()
    api.ios_keychain_remove(account, service)
    click.secho('Successfully removed matching keychain items', fg='green')


def update(args: list) -> None:
    """
        Update matching keychain entry from the keychain

        :param args:
        :return:
    """

    account = _get_flag_value(args, '--account')
    service = _get_flag_value(args, '--service')
    newdata = _get_flag_value(args, '--newdata')

    if not account or not service or not newdata:
        click.secho('All flags need to be set, incl. --account, --service and --newdata')
        return

    click.secho('Updating entries from the iOS keychain...', dim=True)
    click.secho('Account:   {0}'.format(account), dim=True)
    click.secho('Service:   {0}'.format(service), dim=True)
    click.secho('New Data:  {0}'.format(newdata), dim=True)

    api = state_connection.get_api()
    api.ios_keychain_update(account, service, newdata)
    click.secho('Successfully updated matching keychain item', fg='green')


def add(args: list) -> None:
    """
        Adds a new kSecClassGenericPassword keychain entry to the keychain

        :param args:
        :return:
    """

    if not _data_flag_has_identifier(args):
        click.secho('When specifying the --data flag, either --account or '
                    '--service should also be added', fg='red')
        return

    account = _get_flag_value(args, '--account')
    service = _get_flag_value(args, '--service')
    data = _get_flag_value(args, '--data')

    click.secho('Adding a new entry to the iOS keychain...', dim=True)
    click.secho('Account:  {0}'.format(account), dim=True)
    click.secho('Service:  {0}'.format(service), dim=True)
    click.secho('Data:     {0}'.format(data), dim=True)

    api = state_connection.get_api()
    if api.ios_keychain_add(account, service, data):
        click.secho('Successfully added the keychain item', fg='green')
        return

    click.secho('Failed to add the keychain item', fg='red')



import os

import click

from objection.state.connection import state_connection


def clazz(args: list) -> None:
    """
        Simply echoes the source for a generic Hook Manager
        sample for Objective-C hooks with Frida.

        :param args:
        :return:
    """

    js_path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        '../../utils/assets', 'javahookmanager.js'
    )

    with open(js_path, 'r') as f:
        click.secho(f.read(), dim=True)


def simple(args: list) -> None:
    """
        Generate simple hooks for all methods in a Java class.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: android hooking generate simple <class name>', bold=True)
        return

    classname = args[0]

    api = state_connection.get_api()
    methods = api.android_hooking_get_class_methods(classname, False)

    if len(methods) <= 0:
        click.secho('No class / methods found')
        return

    # nasty! :D
    unique_methods = set([x.split('(')[0].split('.')[-1] for x in methods])

    for method in unique_methods:
        hook = """
Java.perform(function() {
    var clazz = Java.use('{clazz}');
    clazz.{method}.implementation = function() {

        //

        return clazz.{method}.apply(this, arguments);
    }
});
    """.replace('{clazz}', classname).replace('{method}', method)

        click.secho(hook, dim=True)



import os
import tempfile
import time

import click
from tabulate import tabulate

from ..state.connection import state_connection
from ..state.device import device_state, Ios, Android
from ..state.filemanager import file_manager_state
from ..utils.helpers import sizeof_fmt

# variable used to cache entries from the ls-like
# commands used in the below helpers. only used
# by the _get_short_*_listing methods.
_ls_cache = {}


def _should_download_folder(args: list) -> bool:
    """
        Checks if --json is in the list of tokens received from the command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--folder' in args


def cd(args: list) -> None:
    """
        Change the current working directory of the device.

        While this method does not actually change any directories,
        it simply updates the value in the file_manager_state property
        that keeps record of the current directory.

        Before changing directories though, some checks are performed
        on the device to at least ensure that the destination directory
        exists.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: cd <destination directory>', bold=True)
        return

    path = args[0]
    current_dir = pwd()

    # nothing to do
    if path == '.':
        return

    # moving one directory back
    device_path_separator = device_state.platform.path_separator

    if path == '..' or path == '..'+device_path_separator:

        split_path = os.path.split(current_dir)

        # nothing to do if we are already at root
        if len(split_path) == 1:
            return

        new_path = ''.join(split_path[:-1])
        click.secho(new_path, fg='green', bold=True)

        file_manager_state.cwd = new_path

        return

    # if we got an absolute path, check if the path
    # actually exists, and then cd to it if we can
    if os.path.isabs(path):

        # assume the path does not exist by default
        does_exist = False

        # normalise path to remove '../'
        if '..'+device_path_separator in path:
            path = os.path.normpath(path).replace('\\', device_path_separator)

        # check for existence based on the runtime
        if device_state.platform == Ios:
            does_exist = _path_exists_ios(path)

        if device_state.platform == Android:
            does_exist = _path_exists_android(path)

        # if we checked with the device that the path exists
        # and it did, update the state manager, otherwise
        # show an error that the path may be invalid
        if does_exist:
            click.secho(path, fg='green', bold=True)

            file_manager_state.cwd = path
            return

        else:
            click.secho('Invalid path: `{0}`'.format(path), fg='red')

    # directory is not absolute, tack it on at the end and
    # see if its legit.
    else:

        proposed_path = device_path_separator.join([current_dir, path])

        # normalise path to remove '../'
        if '..'+device_path_separator in proposed_path:
            proposed_path = os.path.normpath(proposed_path).replace('\\', device_path_separator)
            if proposed_path == '//':
                return

        # assume the proposed_path does not exist by default
        does_exist = False

        # check for existence based on the runtime
        if device_state.platform == Ios:
            does_exist = _path_exists_ios(proposed_path)

        if device_state.platform == Android:
            does_exist = _path_exists_android(proposed_path)

        # if we checked with the device that the path exists
        # and it did, update the state manager, otherwise
        # show an error that the path may be invalid
        if does_exist:
            click.secho(proposed_path, fg='green', bold=True)

            file_manager_state.cwd = proposed_path
            return

        else:
            click.secho('Invalid path: `{0}`'.format(proposed_path), fg='red')


def path_exists(path: str) -> bool:
    """
        Checks if a path exists on remote device.

        :param path:
        :return:
    """

    if device_state.platform == Ios:
        return _path_exists_ios(path)

    if device_state.platform == Android:
        return _path_exists_android(path)


def _path_exists_ios(path: str) -> bool:
    """
        Checks an iOS device if a path exists.

        :param path:
        :return:
    """

    api = state_connection.get_api()
    return api.ios_file_exists(path)


def _path_exists_android(path: str) -> bool:
    """
        Checks an Android device if a path exists.

        :param path:
        :return:
    """

    api = state_connection.get_api()
    return api.android_file_exists(path)


def pwd(args: list = None) -> str:
    """
        Return the current working directory.

        If a record exists in the filemanager state, that directory
        is returned. Else, an environment specific call is made to
        the device to determine the directory it considers itself
        to be working from.

        :param args:
        :return:
    """

    if file_manager_state.cwd is not None:
        return file_manager_state.cwd

    if device_state.platform == Ios:
        return _pwd_ios()

    if device_state.platform == Android:
        return _pwd_android()


def pwd_print(args: list = None) -> None:
    """
        Prints the current working directory.

        :param args:
        :return:
    """

    click.secho('Current directory: {0}'.format(pwd()))


def _pwd_ios() -> str:
    """
        Execute a Frida hook that gets the current working
        directory from an iOS device.

        :return:
    """

    api = state_connection.get_api()
    cwd = api.ios_file_cwd()

    # update the file_manager state's cwd
    file_manager_state.cwd = cwd

    return cwd


def _pwd_android() -> str:
    """
        Execute a Frida hook that gets the current working
        directory from an Android device.

        :return:
    """

    api = state_connection.get_api()
    cwd = api.android_file_cwd()

    # update the file_manager state's cwd
    file_manager_state.cwd = cwd

    return cwd


def ls(args: list) -> None:
    """
        Get a directory listing for a path on a device.
        If no path is provided, the current working directory is used.

        :param args:
        :return:
    """

    # check if we have received a path to ls for.
    if len(args) <= 0:
        path = pwd()
    else:
        path = args[0]
        if not os.path.isabs(path):
            path = device_state.platform.path_separator.join([pwd(), path])

    # based on the runtime, execute the correct ls method.
    if device_state.platform == Ios:
        _ls_ios(path)

    if device_state.platform == Android:
        _ls_android(path)


def _ls_ios(path: str) -> None:
    """
        List files implementation for iOS.

        See:
            http://www.stanford.edu/class/cs193p/cgi-bin/drupal/system/files/lectures/09_Data.pdf

        :param path:
        :return:
    """

    api = state_connection.get_api()
    data = api.ios_file_ls(path)

    def _get_key_if_exists(attribs, key):
        """
            Small helper to grab keys where some may or may
            not exist in the file attributes.

            :param attribs:
            :param key:
            :return:
        """

        if key in attribs:
            return attribs[key]

        return 'n/a'

    def _humanize_size_if_possible(size: str) -> str:
        """
            Small helper method used to 'humanize' file sizes
            if the file size is not recorded as 'n/a'

            :param size:
            :return:
        """

        return sizeof_fmt(int(size)) if size != 'n/a' else 'n/a'

    # if the directory was readable, dump the filesystem listing
    # and attributes to screen.
    click.secho(tabulate(
        [[
            _get_key_if_exists(file_data['attributes'], 'NSFileType').replace('NSFileType', ''),
            _get_key_if_exists(file_data['attributes'], 'NSFilePosixPermissions'),
            _get_key_if_exists(file_data['attributes'], 'NSFileProtectionKey').replace('NSFileProtection', ''),

            # file read / write permissions
            file_data['readable'],
            file_data['writable'],

            # owner name and uid
            _get_key_if_exists(file_data['attributes'], 'NSFileOwnerAccountName') + ' (' +
            _get_key_if_exists(file_data['attributes'], 'NSFileOwnerAccountID') + ')',

            # group name and gid
            _get_key_if_exists(file_data['attributes'], 'NSFileGroupOwnerAccountName') + ' (' +
            _get_key_if_exists(file_data['attributes'], 'NSFileGroupOwnerAccountID') + ')',

            _humanize_size_if_possible(_get_key_if_exists(file_data['attributes'], 'NSFileSize')),
            _get_key_if_exists(file_data['attributes'], 'NSFileCreationDate'),

            file_name,

        ] for file_name, file_data in data['files'].items()], headers=[
            'NSFileType', 'Perms', 'NSFileProtection', 'Read', 'Write', 'Owner', 'Group', 'Size', 'Creation', 'Name'
        ],
    )) if data['readable'] else None

    # handle the permissions summary for this directory
    click.secho('\nReadable: {0}  Writable: {1}'.format(data['readable'], data['writable']), bold=True)


def _ls_android(path: str) -> None:
    """
        Lit files implementation for Android devices.

        :param path:
        :return:
    """

    api = state_connection.get_api()
    data = api.android_file_ls(path)

    def _timestamp_to_str(stamp: str) -> str:
        """
            Small helper method to convert the timestamps we get
            from the Android filesystem to human readable ones.

            :param stamp:
            :return:
        """

        # convert the time to an integer
        stamp = int(stamp)

        if stamp > 0:
            return time.strftime('%Y-%m-%d %H:%M:%S GMT', time.gmtime(stamp / 1000.0))

        return 'n/a'

    click.secho(tabulate(
        [[
            'Directory' if file_data['attributes']['isDirectory'] else 'File',

            _timestamp_to_str(file_data['attributes']['lastModified']),

            # read / write permissions
            file_data['readable'],
            file_data['writable'],
            file_data['attributes']['isHidden'],

            sizeof_fmt(float(file_data['attributes']['size'])),

            file_name,

        ] for file_name, file_data in data['files'].items()], headers=[
            'Type', 'Last Modified', 'Read', 'Write', 'Hidden', 'Size', 'Name'
        ],
    )) if data['readable'] else None

    click.secho('\nReadable: {0}  Writable: {1}'.format(data['readable'], data['writable']), bold=True)


def download(args: list) -> None:
    """
        Downloads a file from a remote filesystem and stores
        it locally.

        This method is simply a proxy to the actual download methods
        used for the appropriate environment.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: filesystem download <remote location> (optional: <local destination>)', bold=True)
        return

    # determine the source and destination file names.
    # if we didnt get a specification of where to dump the file,
    # assume the same name should be used locally.
    source = args[0]
    destination = args[1] if len(args) > 1 else os.path.basename(source)

    should_download_folder = _should_download_folder(args)

    if device_state.platform == Ios:
        _download_ios(source, destination, should_download_folder)

    if device_state.platform == Android:
        _download_android(source, destination, should_download_folder)


def _download_ios(path: str, destination: str, should_download_folder: bool, path_root: bool = True) -> None:
    """
        Download a file from an iOS filesystem and store it locally.

        :param path:
        :param destination:
        :return:
    """

    # if the path we got is not absolute, join it with the
    # current working directory
    if not os.path.isabs(path):
        path = device_state.platform.path_separator.join([pwd(), path])

    api = state_connection.get_api()

    if path_root:
        click.secho('Downloading {0} to {1}'.format(path, destination), fg='green', dim=True)

    if not api.ios_file_readable(path):
        click.secho('Unable to download file. File is not readable.', fg='red')
        return

    if not api.ios_file_path_is_file(path):
        if not should_download_folder:
            click.secho('To download folders, specify --folder.', fg='yellow')
            return

        if os.path.exists(destination):
            click.secho('The target path already exists.', fg='yellow')
            return

        os.makedirs(destination)

        if path_root:
            if not click.confirm('Do you want to download the full directory?', default=True):
                click.secho('Download aborted.', fg='yellow')
                return
            click.secho('Downloading directory recursively...', fg='green')

        data = api.ios_file_ls(path)
        for name, _ in data['files'].items():
            sub_path = device_state.platform.path_separator.join([path, name])
            sub_destination = os.path.join(destination, name)

            _download_ios(sub_path, sub_destination, True, False)
        if path_root:
            click.secho('Recursive download finished.', fg='green')

        return

    if path_root:
        click.secho('Streaming file from device...', dim=True)
    file_data = api.ios_file_download(path)

    if path_root:
        click.secho('Writing bytes to destination...', dim=True)

    with open(destination, 'wb') as fh:
        fh.write(bytearray(file_data['data']))

    click.secho('Successfully downloaded {0} to {1}'.format(path, destination), bold=True)


def _download_android(path: str, destination: str, should_download_folder: bool, path_root: bool = True) -> None:
    """
        Download a file from the Android filesystem and store it locally.

        :param path:
        :param destination:
        :return:
    """

    # if the path we got is not absolute, join it with the
    # current working directory
    if not os.path.isabs(path):
        path = device_state.platform.path_separator.join([pwd(), path])

    api = state_connection.get_api()

    if path_root:
        click.secho('Downloading {0} to {1}'.format(path, destination), fg='green', dim=True)

    if not api.android_file_readable(path):
        click.secho('Unable to download file. Target path is not readable.', fg='red')
        return

    if not api.android_file_path_is_file(path):
        if not should_download_folder:
            click.secho('To download folders, specify --folder.', fg='yellow')
            return

        if os.path.exists(destination):
            click.secho('The target path already exists.', fg='yellow')
            return

        os.makedirs(destination)

        if path_root:
            if not click.confirm('Do you want to download the full directory?', default=True):
                click.secho('Download aborted.', fg='yellow')
                return
            click.secho('Downloading directory recursively...', fg='green')

        data = api.android_file_ls(path)
        for name, _ in data['files'].items():
            sub_path = device_state.platform.path_separator.join([path, name])
            sub_destination = os.path.join(destination, name)

            _download_android(sub_path, sub_destination, True, False)
        if path_root:
            click.secho('Recursive download finished.', fg='green')
        return

    if path_root:
        click.secho('Streaming file from device...', dim=True)
    file_data = api.android_file_download(path)

    if path_root:
        click.secho('Writing bytes to destination...', dim=True)

    with open(destination, 'wb') as fh:
        fh.write(bytearray(file_data['data']))

    click.secho('Successfully downloaded {0} to {1}'.format(path, destination), bold=True)


def upload(args: list) -> None:
    """
        Uploads a local file to the remote operating system.

        This method is just a proxy method to the real upload
        method used based on the runtime that is available.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: filesystem upload <local source> (optional: <remote destination>)', bold=True)
        return

    source = args[0]
    destination = args[1] if len(args) > 1 else device_state.platform.path_separator.join(
        [pwd(), os.path.basename(source)])

    if device_state.platform == Ios:
        _upload_ios(source, destination)

    if device_state.platform == Android:
        _upload_android(source, destination)


def _upload_ios(path: str, destination: str) -> None:
    """
        Upload a file to a remote iOS filesystem.

        :param path:
        :param destination:
        :return:
    """

    if not os.path.isabs(destination):
        destination = device_state.platform.path_separator.join([pwd(), destination])

    api = state_connection.get_api()
    click.secho('Uploading {0} to {1}'.format(path, destination), fg='green', dim=True)

    # if we cant read the file, just stop
    if not api.ios_file_writable(os.path.dirname(destination)):
        click.secho('Unable to upload file. Destination is not writable.', fg='red')
        return

    click.secho('Reading source file...', dim=True)
    with open(path, 'rb') as f:
        data = f.read().hex()

    click.secho('Sending file to device for writing...', dim=True)
    api.ios_file_upload(destination, data)

    click.secho('Uploaded: {0}'.format(destination), dim=True)

    # unset the cache key for this directory so the next short listing
    # will have updated contents
    if os.path.dirname(destination) in _ls_cache:
        del _ls_cache[os.path.dirname(destination)]


def _upload_android(path: str, destination: str) -> None:
    """
        Upload a file to a remote Android filesystem.

        :param path:
        :param destination:
        :return:
    """

    if not os.path.isabs(destination):
        destination = device_state.platform.path_separator.join([pwd(), destination])

    api = state_connection.get_api()
    click.secho('Uploading {0} to {1}'.format(path, destination), fg='green', dim=True)

    # if we cant read the file, just stop
    if not api.android_file_writable(os.path.dirname(destination)):
        click.secho('Unable to upload file. Destination is not writable.', fg='red')
        return

    click.secho('Reading source file...', dim=True)
    with open(path, 'rb') as f:
        data = f.read().hex()

    click.secho('Sending file to device for writing...', dim=True)
    api.android_file_upload(destination, data)

    click.secho('Uploaded: {0}'.format(destination), dim=True)

    # unset the cache key for this directory so the next short listing
    # will have updated contents
    if os.path.dirname(destination) in _ls_cache:
        del _ls_cache[os.path.dirname(destination)]


def rm(args: list) -> None:
    """
        Remove a file from the remote filesystem.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: rm <target remote file>', bold=True)
        return

    target = args[0]

    if not os.path.isabs(target):
        target = device_state.platform.path_separator.join([pwd(), target])

    if not click.confirm('Really delete {0} ?'.format(target)):
        click.secho('Not deleting {0}'.format(target), dim=True)
        return

    if device_state.platform == Ios:
        _rm_ios(target)

    if device_state.platform == Android:
        _rm_android(target)


def _rm_android(t: str) -> None:
    """
        Removes a file from an Android device.

        :param t:
        :return:
    """

    api = state_connection.get_api()

    if not _path_exists_android(t):
        click.secho('{0} does not exist'.format(t), fg='red')
        return

    if api.android_file_delete(t):
        click.secho('{0} successfully deleted'.format(t), fg='green')

    # update the file system cache entry
    if os.path.dirname(t) in _ls_cache:
        del _ls_cache[os.path.dirname(t)]


def _rm_ios(t: str) -> None:
    """
        Removes a file from an iOS device.

        :param t:
        :return:
    """

    api = state_connection.get_api()

    if not _path_exists_ios(t):
        click.secho('{0} does not exist'.format(t), fg='red')
        return

    if api.ios_file_delete(t):
        click.secho('{0} successfully deleted'.format(t), fg='green')

    # update the file system cache entry
    if os.path.dirname(t) in _ls_cache:
        del _ls_cache[os.path.dirname(t)]


def cat(args: list):
    """
        Downloads a file from a remote filesystem and echos
        it's contents

        This method is simply a proxy to the relevant download methods
        that echoes the contents and cleans up after itself.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: filesystem cat <remote location>', bold=True)
        return

    # determine the source and destination file names.
    # if we didnt get a specification of where to dump the file,
    # assume the same name should be used locally.
    source = args[0]
    _, destination = tempfile.mkstemp('.file')

    if device_state.platform == Ios:
        _download_ios(source, destination, False)

    if device_state.platform == Android:
        _download_android(source, destination, False)

    click.secho('====', dim=True)
    with open(destination, 'r', encoding='utf-8', errors='ignore') as f:
        print(f.read(), end='', )
    click.secho('====', dim=True)

    os.remove(destination)


def _get_short_ios_listing() -> list:
    """
        Get a shortened file and directory listing for
        iOS devices.

        :return:
    """

    # default to the pwd. this method is for tab
    # completions anyways.
    directory = pwd()

    # the response for this directory
    resp = []

    # check our cheap cache if we have a listing
    if directory in _ls_cache:
        return _ls_cache[directory]

    api = state_connection.get_api()
    data = api.ios_file_ls(directory)

    # loop the response, marking entries as either being
    # a file or a directory. this response will be stored
    # in the _ls_cache too.
    for name, attribs in data['files'].items():

        # attributes key contains the type
        attributes = attribs['attributes']

        # if the attributes dict does not have the file type,
        # just continue as we cant be sure what it is.
        if 'NSFileType' not in attributes:
            continue

        # append a tuple with name, type
        resp.append((name, 'directory' if attributes['NSFileType'] == 'NSFileTypeDirectory' else 'file'))

    # cache the response so its faster next time!
    _ls_cache[directory] = resp

    # grab the output lets seeeeee
    return resp


def _get_short_android_listing() -> list:
    """
        Get a shortened file and directory listing for
        Android devices.

        :return:
    """

    # default to the pwd. this method is for tab
    # completions anyways.
    directory = pwd()

    # the response for this directory
    resp = []

    # check our cheap cache if we have a listing
    if directory in _ls_cache:
        return _ls_cache[directory]

    api = state_connection.get_api()
    data = api.android_file_ls(directory)

    # loop the response, marking entries as either being
    # a file or a directory. this response will be stored
    # in the _ls_cache too.
    for name, attribs in data['files'].items():
        attributes = attribs['attributes']

        # append a tuple with name, type
        resp.append((name, 'directory' if attributes['isDirectory'] else 'file'))

    # cache the response so its faster next time!
    _ls_cache[directory] = resp

    # grab the output lets seeeeee
    return resp


def list_folders_in_current_fm_directory() -> dict:
    """
        Return folders in the current working directory of the
        Frida attached device.
    """

    resp = {}

    # get the folders based on the runtime
    if device_state.platform == Ios:
        response = _get_short_ios_listing()

    elif device_state.platform == Android:
        response = _get_short_android_listing()

    # looks like we landed in an unknown runtime.
    # just return.
    else:
        return resp

    # loop the response to get entries for the 'directory'
    # type.
    for entry in response:
        file_name, file_type = entry

        if file_type == 'directory':
            if ' ' in file_name:
                resp[f"'{file_name}'"] = file_name
            else:
                resp[file_name] = file_name

    return resp


def list_files_in_current_fm_directory() -> dict:
    """
        Return files in the current working directory of the
        Frida attached device.
    """

    resp = {}

    # check for existence based on the runtime
    if device_state.platform == Ios:
        response = _get_short_ios_listing()

    elif device_state.platform == Android:
        response = _get_short_android_listing()

    # looks like we landed in an unknown runtime.
    # just return.
    else:
        return resp

    # loop the response to get entries for the 'directory'
    # type.
    for entry in response:
        file_name, file_type = entry

        if file_type == 'file':
            if ' ' in file_name:
                resp[f"'{file_name}'"] = file_name
            else:
                resp[file_name] = file_name

    return resp


def list_content_in_current_fm_directory() -> dict:
    """
        Return folders and files in the current working directory of the
        Frida attached device.
    """

    resp = {}

    # check for existence based on the runtime
    if device_state.platform == Ios:
        response = _get_short_ios_listing()

    elif device_state.platform == Android:
        response = _get_short_android_listing()

    # looks like we landed in an unknown runtime.
    # just return.
    else:
        return resp

    # loop the response to get entries.
    for entry in response:
        name, _ = entry

        if ' ' in name:
            resp[f"'{name}'"] = name
        else:
            resp[name] = name

    return resp



import logging
import os
import threading

import click

from .update_checker import check_version


class MakeFileHandler(logging.FileHandler):
    """
        Wrapper Class around the builtin Filehandler.
    """

    def __init__(self, filename: str, mode: str = 'a', encoding: str = None, delay: bool = False) -> None:
        """
            The original FileHandler's init is called, right after the
            directory used to store the objection logfile is created.

            :param filename:
            :param mode:
            :param encoding:
            :param delay:
        """

        os.makedirs(os.path.dirname(filename), exist_ok=True)
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)


def new_secho(text: str, **kwargs) -> None:
    """
        Patch the secho method from the click package so that
        the text that should be echoed is logged first.

        :param text:
        :param kwargs:
        :return:
    """

    logging.info(text)
    real_secho(text, **kwargs)


# Configure the logging used in objection
logger = logging.getLogger()
handler = MakeFileHandler(os.path.expanduser('~/.objection/objection.log'))
formatter = logging.Formatter('%(asctime)s %(levelname)-8s\n%(message)s\n')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)

# monkey patch secho to log to file
real_secho = click.secho
click.secho = new_secho

try:
    # kick off a background thread to check the version of objection
    threading.Thread(target=check_version).start()
except Exception:
    pass



import re
import shlex

import click
from packaging.version import Version
from ..state.app import app_state
from ..state.device import device_state, Ios, Android


def debug_print(message: str) -> None:
    """
        Prints a message if the application is running with
        debugging enabled.

        :param message:
        :return:
    """

    if app_state.should_debug():
        click.secho('[debug] {message}'.format(message=message), dim=True)


def pretty_concat(data: str, at_most: int = 75, left: bool = False) -> str:
    """
        Limits a string to the maximum value of 'at_most',
        ending it off with 3 '.'s. If true is specified for
        the left parameter, the end of the string will be
        used with 3 '.'s prefixed.

        :param data:
        :param at_most:
        :param left:
        :return:

    """

    # do nothing if we are below the max length
    if len(data) <= at_most:
        return data

    if left:
        return '...' + data[len(data) - at_most:]

    return data[:at_most] + '...'


def sizeof_fmt(num: float, suffix: str = 'B') -> str:
    """
        Pretty print bytes
    """

    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return '%3.1f %s%s' % (num, unit, suffix)
        num /= 1024.0
    return '%.1f %s%s' % (num, 'Yi', suffix)


def get_tokens(text: str) -> list:
    """
        Split the text line, shell-style.

        Sometimes we will have strings that don't have the last
        quotes added yet. In those cases, we can just ignore
        shlex errors. :)

        :param text:
        :return:
    """

    try:

        tokens = shlex.split(text)

    except ValueError:

        # return a response that wont match a next command
        tokens = ['lajfhlaksjdfhlaskjfhafsdlkjh']

    return tokens


def clean_argument_flags(args: list) -> list:
    """
        Returns a list of arguments with flags removed.

        Items are considered flags when they are prefixed
        with two dashes.

        :param args:
        :return:
    """

    return [x for x in args if not x.startswith('--')]


def to_snake_case(w: str) -> str:
    """
        https://stackoverflow.com/a/1176023

        :param w:
        :return:
    """

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', w)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def print_frida_connection_help() -> None:
    """
        Prints help information about connecting to devices and
        processes.

        :return:
    """

    click.secho('If you are using a rooted/jailbroken device, specify a process with '
                'the --gadget flag. Eg: objection --gadget "Calendar" explore', fg='red')
    click.secho('If you are using a non rooted/jailbroken device, ensure that your patched application '
                'is running and in the foreground.', fg='red')
    click.secho('')
    click.secho('If you have multiple devices, specify the target device with --serial. A list '
                'of attached device serials can be found with the frida-ls-devices command.', fg='yellow')
    click.secho('')
    click.secho('For more information, please refer to the objection wiki at: '
                'https://github.com/sensepost/objection/wiki', fg='green')


def warn_about_older_operating_systems() -> None:
    """
        Prints a warning to the console about the recommended Android and
        iOS versions to use with objection.

        :return:
    """

    platform = getattr(device_state, 'platform', None)
    version = getattr(device_state, 'version', None)
    if platform is None or version is None:
        return

    android_supported = '5'
    ios_supported = '9'

    # android & ios version warnings
    if platform == Android and (
            Version(version) < Version(android_supported)):
        click.secho('Warning: You appear to be running Android {0} which may result in '
                    'some hooks failing.\nIt is recommended to use at least an Android '
                    'version {1} device with objection.'.format(version, android_supported),
                    fg='yellow')

    # android & ios version warnings
    if platform == Ios and (
            Version(version) < Version(ios_supported)):
        click.secho('Warning: You appear to be running iOS {0} which may result in '
                    'some hooks failing.\nIt is recommended to use at least an iOS '
                    'version {1} device with objection.'.format(version, ios_supported),
                    fg='yellow')



import click

from objection.state.connection import state_connection


def android_proxy_set(args: list = None) -> None:
    """
        Sets a proxy specifically within the application.

        :param args:
        :return:
    """

    if len(args) != 2:
        click.secho('Usage: android proxy set <ip address> <port>', bold=True)
        return

    api = state_connection.get_api()
    api.android_proxy_set(args[0], args[1])



from flask import Blueprint, jsonify, request, abort

from objection.state.connection import state_connection

bp = Blueprint('script', __name__, url_prefix='/script')


@bp.route('/runonce', methods=('POST',))
def runonce():
    """
        Run an arbitrary script in the connected frida
        enabled device.

        Responses are JSON encoded by default, but can be raw by adding
        ?json=false as a query string parameter.

        :return:
    """

    source = request.data.decode('utf-8')

    if len(source) <= 0:
        return abort(jsonify(message='Missing or empty script received'))

    try:

        # run the script
        response = state_connection.get_agent().single(source)

        if 'json' in request.args and request.args.get('json').lower() == 'false':
            return response

    except Exception as e:
        return abort(jsonify(message='Script failed to run: {e}'.format(e=str(e))))

    return jsonify(response)



import collections

from prompt_toolkit.completion import Completer, Completion, CompleteEvent
from prompt_toolkit.document import Document

from .commands import COMMANDS
from ..utils.helpers import get_tokens


class CommandCompleter(Completer):
    """
        The objection REPL command completer.
    """

    def __init__(self) -> None:
        super(CommandCompleter, self).__init__()
        self.COMMANDS = COMMANDS

    def find_completions(self, document: Document) -> dict:
        """
            Find tab completions from the commands repository.

            Completions are returned based on tokens extracted
            from the command text received by prompt_toolkit. A
            dictionary is then walked, matching a token to a
            nested dictionary until no more dictionaries are
            available. The resultant dictionary then becomes
            the suggestions for tab completion.

            Some commands may have 'dynamic' completions, such as
            file system related commands. They are defined with a
            'dynamic' key, and the method defined as the value for
            this key is executed to get completions.

            :param document:
            :return:
        """

        # extract tokens from the document similar to
        # how a shell invocation would have been done.
        # we will also cleanup flags that come in the form
        #  of --flag so that multiples can be suggested.
        tokens = [token for token in get_tokens(document.text) if not token.startswith('--')]

        # extract the flags in the received tokens. This list
        # will be used to remove suggested flags from those
        # already present in the command.
        flags = [flag for flag in get_tokens(document.text) if flag.startswith('--')]

        # start with the current suggestions dictionary being
        # all commands
        current_suggestions = self.COMMANDS

        # when the tokens are extracted, we are expecting something in
        # the format of:
        #   command sub_command sub_sub_command
        # so, lets use that and search the the COMMAND dictionary for
        # the last dictionary with a correct suggestion
        for token in tokens:

            candidate = token.lower()

            if candidate in list(current_suggestions.keys()):

                # if there are sub commands, grab them
                if 'commands' in current_suggestions[candidate]:
                    current_suggestions = current_suggestions[candidate]['commands']

                # dynamic commands change based on the current status of the
                # environment, so, call the method defined
                elif 'dynamic' in current_suggestions[candidate]:
                    current_suggestions = current_suggestions[candidate]['dynamic']()

                # make --flags in the 'flags' key tab completable
                elif 'flags' in current_suggestions[candidate]:
                    current_suggestions = {
                        flag: '' for flag in current_suggestions[candidate]['flags'] if flag not in flags
                    }

                # in this case, there are probably no sub commands, so return
                # an empty dictionary
                else:
                    return {}

        suggestions = {}

        # once we have the deepest suggestions dictionary in the
        # current_suggestions variable, loop through and check for
        # 'sorta' matched versions
        if current_suggestions and len(current_suggestions) > 0:
            for k, _ in current_suggestions.items():

                # fuzzy-ish matching when part of a word is in a suggestion
                if document.get_word_before_cursor().lower() in k.lower():
                    suggestions[k] = current_suggestions[k]

        return suggestions

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Completion:
        """
            The main method that gets called by prompt-toolkit to
            determine which completions to show. This

            :param document:
            :param complete_event:
            :return:
        """

        # get the stuff we have typed so far
        word_before_cursor = document.get_word_before_cursor()

        # if this is an os command, we can't complete anything
        if document.text.startswith('!'):
            return

        commands = self.find_completions(document)

        # if there are no commands return
        if len(commands) <= 0:
            return

        # sort alphabetically
        commands = collections.OrderedDict(sorted(list(commands.items()), key=lambda t: t[0]))

        # loop the commands that we have determined to be useful
        # based on the current text input and populate a 'meta' field
        # if one exists.
        for cmd, extra in commands.items():
            meta = extra['meta'] if type(extra) is dict and 'meta' in extra else None

            # finally, yield the generator for completions
            yield Completion(cmd, -(len(word_before_cursor)), display_meta=meta)



import click

from ..commands.filemanager import pwd
from ..state.connection import state_connection


def start(args: list) -> None:
    """
        Start's an http server, exposing the mobile devices filesystem.

        :param args:
        :return:
    """

    port = 9000

    if len(args) > 0:
        port = int(args[0])

    click.secho('Starting server on port {port}...'.format(port=port), dim=True)

    api = state_connection.get_api()
    api.http_server_start(pwd(), port)


def stop(args: list) -> None:
    """
        Stops the on device HTTP server

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.http_server_stop()


def status(args: list) -> None:
    """
        Get the status of the HTTP server

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.http_server_status()



import atexit
from random import randint

import click
import frida

from objection.state.connection import state_connection


class Job(object):
    """  A class representing a REPL Job or agent Job with one or more hooks. """

    def __init__(self, name, job_type, handle, uuid: int = None) -> None:
        """
            Init a new job. This requires the job_type to know how to manage the job as well as a handle
            to manage and kill the job.

            :param name:
            :param job_type:
            :param handle:
            :return:
        """
        if uuid is not None:
            self.uuid = int(uuid)
        else:
            self.uuid = randint(100000, 999999)
        self.name = name
        self.job_type = job_type
        self.handle = handle

    def end(self):
        """
            Revert hooks that the job created.

            :return:
        """
        if self.job_type == "script":

            click.secho("[job manager] Killing job {0}. Name: {1}. Type: {2}"
                        .format(self.uuid, self.name, self.job_type), dim=True)
            self.handle.unload()
        elif self.job_type == "hook":
            api = state_connection.get_api()
            api.jobs_kill(self.uuid)
        else:
            click.secho(('[job {0}] - Unknown job type {1}'.format(self.uuid, self.job_type)), fg='red', dim=True)


class JobManagerState(object):
    """  A class representing the current Job manager. """

    def __init__(self) -> None:
        """
            Init a new job state manager. This method will also
            register an atexit(), ensuring that cleanup operations
            are performed on jobs when this class is GC'd.
        """

        self.jobs: dict[int, Job] = {}

        atexit.register(self.cleanup)

    def add_job(self, new_job: Job) -> None:
        """
            Adds a job to the job state manager.

            :param new_job:
            :return:
        """

        # avoid duplicate jobs.
        if new_job.uuid not in self.jobs:
            self.jobs[new_job.uuid] = new_job

    def remove_job(self, job_uuid: int):
        """
            Removes a job from the job state manager.

            :param job_uuid:
            :return Job:
        """
        if job_uuid not in self.jobs:
            click.secho(f"Error: Job with ID {job_uuid} does not exist.", fg='red')
            return

        job_to_remove = self.jobs.pop(job_uuid)
        job_to_remove.end()

    def cleanup(self) -> None:
        """
            Clean up all the jobs in the job manager.

            This method is typical called when at the end of an
            objection session.

            :return:
        """

        for uuid in list(self.jobs.keys()):
            try:
                job = self.jobs.pop(uuid)
                job.end()

            except frida.InvalidOperationError:
                click.secho(('[job manager] Job: {0} - An error occurred stopping job. Device may '
                             'no longer be available.'.format(uuid)), fg='red', dim=True)


job_manager_state = JobManagerState()



import argparse
import atexit
import json
import sys
from dataclasses import dataclass
from pathlib import Path
from pprint import pprint

import click
import frida

from objection.state.app import app_state
from objection.state.connection import state_connection
from objection.state.device import device_state, Ios, Android
from objection.state.jobs import job_manager_state, Job
from objection.utils.helpers import debug_print


@dataclass
class AgentConfig(object):
    """ Default configuration for an Agent instance """

    name: str
    host: str = None
    port: int = None
    device_type: str = 'usb'
    device_id: str = None
    foremost: bool = False
    spawn: bool = False
    pause: bool = True
    debugger: bool = False
    uid: int = None


class OutputHandlers(object):
    """ Output handlers for an Agent instance """

    def device_output(self):
        pass

    def device_lost(self):
        pass

    @staticmethod
    def session_on_detached(message: str, crash):
        """
            The callback to run for the detach signal

            :param message:
            :param crash:

            :return:
        """

        try:

            # log the hook response if needed
            if app_state.should_debug():
                click.secho('- [incoming message] ' + '-' * 18, dim=True)
                click.secho(json.dumps(message, indent=2, sort_keys=True), dim=True)
                click.secho('- [./incoming message] ' + '-' * 16, dim=True)

            # process the response
            if message:
                click.secho('(session detach message) ' + message, fg='red')

            # Frida 12.3 crash reporting
            # https://www.nowsecure.com/blog/2019/02/07/frida-12-3-debuts-new-crash-reporting-feature/
            if crash:
                click.secho('(process crash report)', fg='red')
                click.secho('\n\t{0}'.format(crash.report), dim=True)

        except Exception as e:
            click.secho('Failed to process an incoming message for a session detach signal: {0}'.format(e), fg='red',
                        bold=True)
            raise e

    @staticmethod
    def script_on_message(message: dict, data):
        """
            The callback to run when a message is received from the agent.

            :param message:
            :param data:

            :return:
        """

        try:
            # log the hook response if needed
            if app_state.should_debug():
                click.secho('- [incoming message] ' + '-' * 18, dim=True)
                click.secho(json.dumps(message, indent=2, sort_keys=True), dim=True)
                click.secho('- [./incoming message] ' + '-' * 16, dim=True)

            # process the response
            if message and 'payload' in message:
                if len(message['payload']) > 0:
                    if isinstance(message['payload'], dict):
                        click.secho('(agent) ' + json.dumps(message['payload']))
                    elif isinstance(message['payload'], str):
                        click.secho('(agent) ' + message['payload'])
                    else:
                        click.secho('Dumping unknown agent message', fg='yellow')
                        pprint(message['payload'])

        except Exception as e:
            click.secho('Failed to process an incoming message from agent: {0}'.format(e), fg='red', bold=True)
            raise e


class Agent(object):
    """ Class to manage the lifecycle of the objection Frida agent """

    agent_path: Path = None
    c: AgentConfig

    handlers: OutputHandlers

    device: frida.core.Device = None
    session: frida.core.Session = None
    script: frida.core.Script = None

    pid: int = None
    resumed: bool = True

    def __init__(self, config: AgentConfig):
        """ initialises the agent class """

        self.agent_path = Path(__file__).parent.parent / 'agent.js'
        if not self.agent_path.exists():
            raise Exception(f'Unable to locate Objection agent sources at: {self.agent_path}. '
                            'If this is a development install, check the wiki for more '
                            'information on building the agent.')
        debug_print('Agent path is: {path}'.format(path=self.agent_path))

        self.config = config
        debug_print(f'agent config: {self.config}')
        self.handlers = OutputHandlers()

        atexit.register(self.teardown)

    def _get_agent_source(self) -> str:
        """
            Loads the frida-compiled agent from disk.

            :return:
        """

        with open(self.agent_path, 'r', encoding='utf-8') as f:
            src = f.readlines()

        return ''.join([str(x) for x in src])

    def set_device(self):
        """
            Set's the target device to work with.

            :return:
        """

        if self.config.device_id is not None:
            self.device = frida.get_device(self.config.device_id)

        elif (self.config.host is not None) or (self.config.device_type == 'remote'):
            if self.config.host is None:
                self.device = frida.get_remote_device()
            else:
                host = self.config.host
                port = self.config.port
                self.device = frida.get_device_manager() \
                    .add_remote_device(f'{host}:{port}' if host is not None else f'127.0.0.1:{port}')

        elif self.config.device_type is not None:
            for dev in frida.enumerate_devices():
                if dev.type == self.config.device_type:
                    self.device = dev
        else:
            self.device = frida.get_local_device()

        # surely we have a device by now?
        if self.device is None:
            raise Exception('Unable to find a device')

        self.device.on('output', self.handlers.device_output)
        self.device.on('lost', self.handlers.device_lost)

        debug_print(f'device determined as: {self.device}')

    def set_target_pid(self):
        """
            Set's the PID we should attach to. This method will spawn the
            target if needed. The resumed value is also toggled here.

            Defaults:
                resumed: bool = True

            :return:
        """

        if (self.config.name is None) and (not self.config.foremost):
            click.secho('Need a target name to spawn/attach to', fg='red')
            sys.exit(1)

        if self.config.foremost:
            try:
                app = self.device.get_frontmost_application()
            except Exception as e:
                click.secho(f'Could not get foremost application on {self.device.name}: {e}', fg='red')
                sys.exit(1)

            if app is None:
                click.secho(f'No foremost application on {self.device.name}', fg='red')
                sys.exit(1)

            self.pid = app.pid
            # update the global state for the prompt etc.
            state_connection.name = app.identifier

        elif self.config.spawn:
            if self.config.uid is not None:
                if self.device.query_system_parameters()['os']['id'] != 'android':
                    raise Exception('--uid flag can only be used on Android.')
                self.pid = self.device.spawn(self.config.name, uid=int(self.config.uid))
            else:
                try:
                    self.pid = self.device.spawn(self.config.name)
                except frida.InvalidArgumentError:
                    pass

                # Maybe we have an app name and not identifier
                app_list = self.device.enumerate_applications()
                app_name_lc = self.config.name.lower()

                matching_app = [app for app in app_list if app.name.lower() == app_name_lc]
                # Don't care about matching_app[0].pid not in (0, None), if already running we restart anyway.
                if len(matching_app) == 1:
                    debug_print("Found app by name instead of package, spawning.")
                    self.pid = self.device.spawn(matching_app[0].identifier)

            self.resumed = False
        else:
            # check if the name is actually an integer. this way we can
            # assume we got the target PID already
            try:
                self.pid = int(self.config.name)
            except ValueError:
                pass

            # maybe we have a process name
            if self.pid is None:
                try:
                    self.pid = self.device.get_process(self.config.name).pid
                except frida.ProcessNotFoundError:
                    pass

            if self.pid is None:
                # maybe we have an app identifier/package name
                app_list = self.device.enumerate_applications()
                app_name_lc = self.config.name.lower()
                matching_app = [app for app in app_list if app.identifier.lower() == app_name_lc]
                if len(matching_app) == 1 and matching_app[0].pid not in (0, None):
                    debug_print("Found app by package name.")
                    self.pid = matching_app[0].pid
                elif len(matching_app) > 1:
                    app_list_str = ', '.join([f"{app.identifier}: {app.pid}" for app in matching_app])
                    click.secho("Ambiguous identifier. Applications with the same identifier found: "+ app_list_str, fg='red')
                    sys.exit(1)

        if self.pid is None:
            click.secho("Unable to find target application.", fg='red', bold=True)
            sys.exit(1)

        debug_print(f'process PID determined as {self.pid}')

    def attach(self):
        """
            Attaches to an enumerated PID, injecting the objection agent.

            :return:
        """

        if self.pid is None:
            raise Exception('A PID needs to be set before attach()')

        if self.config.uid is None:
            debug_print(f'Attaching to PID: {self.pid}')
            self.session = self.device.attach(self.pid)
        else:
            self.session = self.device.attach(self.pid)

        self.session.on('detached', self.handlers.session_on_detached)

        if self.config.debugger:
            click.secho('debugger enabled and runtime set to v8. visit chrome://inspect', bold=True)
            self.script = self.session.create_script(source=self._get_agent_source(), runtime='v8')
            self.script.enable_debugger()
        else:
            self.script = self.session.create_script(source=self._get_agent_source())

        self.script.on('message', self.handlers.script_on_message)
        self.script.load()

    def attach_script(self, job_name, source):
        """
            Attaches an arbitrary script session.

            :param job_name:
            :param source:
            :return:
        """

        session: frida.core.Session = self.device.attach(self.pid)
        script: frida.core.Script = session.create_script(source=source)
        script.on('message', self.handlers.script_on_message)
        script.load()

        script_job = Job(job_name, 'script', script)
        job_manager_state.add_job(script_job)

    def update_device_state(self):
        """
            Updates the device_state. Useful in other parts where we
            need platform specific decisions.

            :return:
        """

        params = self.device.query_system_parameters()

        # set os platform
        if params['os']['id'] == 'ios':
            device_state.set_platform(Ios)
        elif params['os']['id'] == 'android':
            device_state.set_platform(Android)
        else:
            rt = self.exports().env_runtime()
            if rt == 'ios':
                device_state.set_platform(Ios)
            elif rt == 'android':
                device_state.set_platform(Android)                

        # set os version
        device_state.set_version(params['os']['version'])

    def resume(self):
        """
            Resume the target pid.

            :return:
        """

        if self.resumed:
            return

        if not self.pid:
            raise Exception('Cannot resume without self.pid')

        self.device.resume(self.pid)
        self.resumed = True

    def exports(self):
        """
            Returns the RPC exports exposed by the Frida agent

            :return:
        """

        if not self.script:
            raise Exception('Need a script created before reading exports()')

        return self.script.exports_sync

    def run(self):
        """
            Run the Agent by getting a device, setting the target pid and attaching.
            If we should skip pausing, also resume the process.

            :return:
        """

        self.set_device()
        self.set_target_pid()
        self.attach()

        # internal state
        self.update_device_state()

        if not self.config.pause:
            debug_print('asked to run without pausing, so resuming in run()')
            self.resume()

    def teardown(self):
        try:
            if self.script:
                click.secho('Asking jobs to stop...', dim=True)
                job_manager_state.cleanup()
                click.secho('Unloading objection agent...', dim=True)
                self.script.unload()
        except frida.InvalidOperationError as e:
            click.secho(f'Unable to run cleanups: {e}', fg='yellow', dim=True)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('name', help='target app to attach/spawn. needs to be bundle '
                                     'identifier for spawn')
    parser.add_argument('--no-spawn', dest='no_spawn', default=True, action='store_false',
                        help='do not try and spawn the target app')
    parser.add_argument('--no-pause', dest='no_pause', default=True, action='store_false',
                        help='resume the app after spawning')
    parser.add_argument('--debug', default=False, action='store_true', help='print debug logging')
    args = parser.parse_args()

    if args.name is None:
        print('error: need a target app to attach/spawn')
        sys.exit(1)

    if args.debug:
        app_state.debug = True

    c = AgentConfig(name=args.name, spawn=args.no_spawn, pause=args.no_pause)
    a = Agent(config=c)
    a.run()

    print(a.exports().env_frida())



from objection.state.connection import state_connection


def disable(args: list = None) -> None:
    """
        Attempts to disable jailbreak detection.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.ios_jailbreak_disable()


def simulate(args: list = None) -> None:
    """
        Attempts to simulate a Jailbroken environment

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.ios_jailbreak_enable()



import importlib.util
import os
import traceback
import uuid

import click

from ..utils.plugin import Plugin as PluginType


def load_plugin(args: list = None) -> None:
    """
        Loads an objection plugin.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: plugin load <plugin path> (<plugin namespace>)', bold=True)
        return

    path = os.path.abspath(args[0])
    if os.path.isdir(path):
        path = os.path.join(path, '__init__.py')

    if not os.path.exists(path):
        click.secho('[plugin] {0} does not appear to be a valid plugin. Missing __init__.py'.format(
            os.path.dirname(path)), fg='red', dim=True)
        return

    spec = importlib.util.spec_from_file_location(str(uuid.uuid4())[:8], path)
    plugin = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(plugin)

    namespace = plugin.namespace
    if len(args) >= 2:
        namespace = args[1]

    plugin.__name__ = namespace

    # try and load the plugin (aka: run its __init__)
    try:

        instance = plugin.plugin(namespace)
        assert isinstance(instance, PluginType)

    except AssertionError:
        click.secho('Failed to load plugin \'{0}\'. Invalid plugin type.'.format(namespace), fg='red', bold=True)
        return

    except Exception as e:
        click.secho('Failed to load plugin \'{0}\' with error: {1}'.format(namespace, str(e)), fg='red', bold=True)
        click.secho('{0}'.format(traceback.format_exc()), dim=True)
        return

    from ..console import commands
    commands.COMMANDS['plugin']['commands'][instance.namespace] = instance.implementation
    click.secho('Loaded plugin: {0}'.format(plugin.__name__), bold=True)



import json

import click
from tabulate import tabulate

from objection.state.connection import state_connection


def _should_dump_json(args: list) -> bool:
    """
        Check if --json is part of the arguments.

        :param args:
        :return:
    """

    return '--json' in args


def get(args: list) -> None:
    """
        Gets cookies using the iOS NSHTTPCookieStorage sharedHTTPCookieStorage
        and prints them to the screen.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    cookies = api.ios_cookies_get()

    if _should_dump_json(args):
        print(json.dumps(cookies, indent=4))
        return

    if len(cookies) <= 0:
        click.secho('No cookies found')
        return

    click.secho(tabulate(
        [[
            cookie['name'],
            cookie['value'],
            cookie['expiresDate'],
            cookie['domain'],
            cookie['path'],
            cookie['isSecure'],
            cookie['isHTTPOnly']
        ] for cookie in cookies], headers=['Name', 'Value', 'Expires', 'Domain', 'Path', 'Secure', 'HTTPOnly'],
    ))



import requests


class Github(object):
    """ Interact with Github """

    GITHUB_LATEST_RELEASE = 'https://api.github.com/repos/frida/frida/releases/latest'
    GITHUB_TAGGED_RELEASE = 'https://api.github.com/repos/frida/frida/releases/tags/{tag}'

    # the 'context' of this Github instance
    gadget_version = None

    def __init__(self, gadget_version: str = None):
        """
            Init a new instance of Github
        """

        if gadget_version:
            self.gadget_version = gadget_version

        self.request_cache = {}

    def _call(self, endpoint: str) -> dict:
        """
            Make a call to Github and cache the response.

            :param endpoint:
            :return:
        """

        # return a cached response if possible
        if endpoint in self.request_cache:
            return self.request_cache[endpoint]

        # get a new response
        results = requests.get(endpoint).json()

        # cache it
        self.request_cache[endpoint] = results

        # and return it
        return results

    def get_latest_version(self) -> str:
        """
            Call Github and get the tag_name of the latest
            release.

            :return:
        """

        self.gadget_version = self._call(self.GITHUB_LATEST_RELEASE)['tag_name']

        return self.gadget_version

    def get_assets(self) -> dict:
        """
            Gets the assets for the currently selected gadget_version.

            :return:
        """

        assets = self._call(self.GITHUB_TAGGED_RELEASE.format(tag=self.gadget_version))

        if 'assets' not in assets:
            raise Exception(('Unable to determine assets for gadget version \'{0}\'. '
                             'Are you sure this version is available on Github?').format(self.gadget_version))

        return assets['assets']



import json
from typing import Optional

import click

from objection.state.connection import state_connection
from objection.utils.helpers import clean_argument_flags


def _is_pattern_or_constant(s: str) -> bool:
    """
        Check if a provided pattern matches "CLASS!METHOD"

        :param s:
        :return:
    """

    # No pattern case
    if "!" not in s:
        return True

    # Check if CLASS and METHOD is defined at all
    parts = s.split('!')
    if len(parts) != 2:
        return False
    elif len(parts[0]) == 0 or len(parts[1]) == 0:
        return False

    return True


def _string_is_true(s: str) -> bool:
    """
        Check if a string should be considered as "True"

        :param s:
        :return:
    """

    return s.lower() in ('true', 'yes')


def _should_dump_backtrace(args: list = None) -> bool:
    """
        Check if --dump-backtrace is part of the arguments.

        :param args:
        :return:
    """

    return '--dump-backtrace' in args


def _should_watch(args: list = None) -> bool:
    """
        Check if --dump-args is part of the arguments.

        :param args:
        :return:
    """

    return '--watch' in args


def _should_dump_args(args: list = None) -> bool:
    """
        Check if --dump-args is part of the arguments.

        :param args:
        :return:
    """

    return '--dump-args' in args


def _should_dump_return_value(args: list = None) -> bool:
    """
        Check if --dump-return is part of the arguments.

        :param args:
        :return:
    """

    return '--dump-return' in args


def _should_dump_json(args: list) -> bool:
    """
        Check if --json is part of the arguments.

        :param args:
        :return:
    """

    return '--json' in args


def _should_be_quiet(args: list) -> bool:
    """
        Check if --quiet is part of the arguments.

        :param args:
        :return:
    """

    return '--quiet' in args


def _should_print_only_classes(args: list = None) -> bool:
    """
        Check if --only-classes is part of the arguments.

        :param args:
        :return:
    """

    return '--only-classes' in args


def _get_flag_value(flag: str, args: list) -> Optional[str]:
    """
        Gets the value for a flag

        :param flag:
        :param args:
        :return:
    """

    target = None

    for i in range(len(args)):
        if args[i] == flag:
            target = i + 1

    if target is None:
        return None
    elif target < len(args):
        return args[target]
    else:
        return None


def show_android_classes(args: list = None) -> None:
    """
        Show the currently loaded classes. 
        Note that Java classes are only loaded when they are used, 
        so not all classes may be present.

        :return:
    """

    api = state_connection.get_api()
    classes = api.android_hooking_get_classes()

    # print the enumerated classes
    for class_name in sorted(classes):
        click.secho(class_name)

    click.secho('\nFound {0} classes'.format(len(classes)), bold=True)


def show_android_class_loaders(args: list = None) -> None:
    """
        Show the currently registered class loaders.

        :return:
    """

    api = state_connection.get_api()
    loaders = api.android_hooking_get_class_loaders()

    # print the enumerated classes
    for loader in sorted(loaders):
        click.secho('* {0}'.format(loader))

    click.secho('\nFound {0} class loaders'.format(len(loaders)), bold=True)


def show_android_class_methods(args: list = None) -> None:
    """
        Shows the methods available on an Android class.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: android hooking list class_methods <class name>', bold=True)
        return

    class_name = args[0]

    api = state_connection.get_api()
    methods = api.android_hooking_get_class_methods(class_name)

    # print the enumerated classes
    for class_name in sorted(methods):
        click.secho(class_name)

    click.secho('\nFound {0} method(s)'.format(len(methods)), bold=True)


def notify(args: list = None) -> None:
    """
        Notify when a class becomes available.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: android hooking notify <pattern>', bold=True)
        return

    query = args[0]
    if not _is_pattern_or_constant(query):
        click.secho('Incorrect query syntax, please use <class>!<method> or just the class name', fg='red')
        return

    api = state_connection.get_api()
    should_watch = _should_watch(args)
    dump_arguments = _should_dump_args(args)
    dump_backtrace = _should_dump_backtrace(args)
    dump_return = _should_dump_return_value(args)
    api.android_hooking_lazy_watch_for_pattern(query, 
        should_watch, dump_arguments, 
        dump_return, 
        dump_backtrace)


def watch(args: list = None) -> None:
    """
        Hook functions and print useful information when they are called.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 1:
        click.secho('Usage: android hooking watch <package pattern> '
                    '(eg: com.example.test, *com.example*!*, com.example.test!toString)'
                    '(optional: --dump-args) '
                    '(optional: --dump-backtrace) '
                    '(optional: --dump-return)',
                    bold=True)
        return

    query = args[0]
    if not _is_pattern_or_constant(query):
        click.secho('Incorrect query syntax, please use <CLASS>!<METHOD>', fg='red')
        return

    api = state_connection.get_api()
    api.android_hooking_watch(query,
                              _should_dump_args(args),
                              _should_dump_backtrace(args),
                              _should_dump_return_value(args))
    return


def search(args: list = None) -> None:
    """
        Enumerates the current Android application for classes and methods.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: android hooking search \'<class>!<method>\n\''
                    '(optional: --json <filename>)'
                    '(optional: --only-classes)', bold=True)
        return

    query = args[0]

    if not _is_pattern_or_constant(query):
        click.secho('Incorrect query syntax, please use <class>!<method>', fg='red')
        return

    api = state_connection.get_api()
    results = api.android_hooking_enumerate(query)

    # Only get overloads if this flag is specified, otherwise just enumerating can be kind of slow
    if _should_dump_json(args):
        results_json = {
            'meta': {
                'runtime': 'java'
            }
        }

        for result in results:
            for _class in result['classes']:
                loader = result['loader']
                if loader is not None:
                    # <instance: java.lang.ClassLoader, $className: dalvik.system.PathClassLoader>
                    # but we only care about the className
                    start_index = loader.find('$className: ') + 12
                    start_part = loader[start_index:]
                    if start_part.find('>'):
                        end_index = start_part.find('>')
                    else:
                        end_index = start_part.find(' ')
                    loader = start_part[:end_index]

                _class['overloads'] = api.android_hooking_get_class_methods_overloads(_class['name'], _class['methods'],
                                                                                      loader)

        target_file = _get_flag_value('--json', args)
        if target_file:
            results_json['data'] = results
            with open(target_file, 'w') as fd:
                fd.write(json.dumps(results_json))
                click.secho(f'JSON dumped to file {target_file}', bold=True)

        return

    # just print to the console
    for result in results:
        for _class in result['classes']:
            if _should_print_only_classes(args):
                print(_class['name'])
                continue

            for method in _class['methods']:
                print(f'{_class["name"]}.{method}')


def show_registered_broadcast_receivers(args: list = None) -> None:
    """
        Enumerate all registered BroadcastReceivers

        :return:
    """

    api = state_connection.get_api()
    receivers = api.android_hooking_list_broadcast_receivers()

    for class_name in sorted(receivers):
        click.secho(class_name)

    click.secho('\nFound {0} classes'.format(len(receivers)), bold=True)


def show_registered_services(args: list = None) -> None:
    """
        Enumerate all registered Services

        :return:
    """

    api = state_connection.get_api()
    services = api.android_hooking_list_services()

    for class_name in sorted(services):
        click.secho(class_name)

    click.secho('\nFound {0} classes'.format(len(services)), bold=True)


def show_registered_activities(args: list = None) -> None:
    """
        Enumerate all registered Activities

        :return:
    """

    api = state_connection.get_api()
    activities = api.android_hooking_list_activities()

    for class_name in sorted(activities):
        click.secho(class_name)

    click.secho('\nFound {0} classes'.format(len(activities)), bold=True)


def get_current_activity(args: list = None) -> None:
    """
        Get the currently active activity

        :return:
    """

    api = state_connection.get_api()
    activity = api.android_hooking_get_current_activity()

    click.secho('Activity: {0}'.format(activity['activity']), bold=True)
    click.secho('Fragment: {0}'.format(activity['fragment']))


def set_method_return_value(args: list = None) -> None:
    """
        Sets a Java methods return value to a specified boolean.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) < 2:
        click.secho(('Usage: android hooking set return_value '
                     '"<fully qualified class method>" "<optional overload>" (eg: "com.example.test.doLogin") '
                     '<true/false>'),
                    bold=True)
        return

    # make sure we got a true/false
    if args[-1].lower() not in ('true', 'false'):
        click.secho('Return value must be set to either true or false', bold=True)
        return

    class_name = args[0].replace('\'', '"')  # fun!

    # check if we got an overload
    overload_filter = args[1].replace(' ', '') if len(args) == 3 else None
    retval = True if _string_is_true(args[-1]) else False

    api = state_connection.get_api()
    api.android_hooking_set_method_return(class_name,
                                          overload_filter,
                                          retval)



import pprint

import click
from prompt_toolkit import prompt
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.javascript import JavascriptLexer
from tabulate import tabulate

from objection.state.connection import state_connection


def _should_ignore_methods_with_arguments(args) -> bool:
    """
        Check if the --without-arguments flag exists

        :param args:
        :return:
    """

    return len(args) > 0 and '--without-arguments' in args


def _should_return_as_string(args) -> bool:
    """
        Check if the --return-string flag exists

        :param args:
        :return:
    """

    return len(args) > 0 and '--return-string' in args


def instances(args: list) -> None:
    """
        Asks the agent to print the currently live instances of a particular class

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: android heap search instances <class> (eg: com.example.test)', bold=True)
        return

    target_class = args[0]

    api = state_connection.get_api()
    instance_results = api.android_heap_get_live_class_instances(target_class)

    if len(instance_results) <= 0:
        return

    click.secho(tabulate(
        [[
            entry['hashcode'],
            entry['classname'],
            entry['tostring'],
        ] for entry in instance_results], headers=['Hashcode', 'Class', 'toString()'],
    ))


def methods(args: list) -> None:
    """
        Get the methods available on a handle

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: android heap print methods <hashcode> (eg: 24688232)', bold=True)
        return

    target_handle = int(args[0])

    api = state_connection.get_api()
    method_results = api.android_heap_print_methods(target_handle)

    # apply argument filters
    # we assume methods that end with braces don't need arguments
    if _should_ignore_methods_with_arguments(args):
        method_results[1] = list(filter(lambda x: '()' in x, method_results[1]))

    click.secho(tabulate(
        [[
            entry,
        ] for entry in method_results], headers=['Method'],
    ))


def execute(args: list) -> None:
    """
        Executes a method on a handle which is assumed to be a Java
        class instance.

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: android heap execute method <hashcode> <method> (eg: 24688232)', bold=True)
        return

    target_handle = int(args[0])
    method = args[1]

    api = state_connection.get_api()
    exec_results = api.android_heap_execute_handle_method(target_handle, method,
                                                          _should_return_as_string(args))

    if exec_results:
        if isinstance(exec_results, dict):
            click.secho(pprint.pformat(exec_results))
        else:
            click.secho(str(exec_results))


def fields(args: list) -> None:
    """

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: android heap print fields <hashcode> (eg: 24688232)', bold=True)
        return

    target_handle = int(args[0])

    api = state_connection.get_api()
    field_results = api.android_heap_print_fields(target_handle)

    click.secho(tabulate(
        [[
            value['name'],
            value['value']
        ] for value in field_results], headers=['Name', 'Value'],
    ))


def evaluate(args: list) -> None:
    """
        Evaluates JavaScript on a handle

        :param args:
        :return:
    """

    if len(args) < 1:
        click.secho('Usage: android heap execute js <hashcode> (eg: 24688232)', bold=True)
        return

    target_handle = int(args[0])

    js = prompt(
        click.secho('(The hashcode at `{handle}` will be available as the `clazz` variable.)'.format(
            handle=target_handle
        ), dim=True),
        multiline=True, lexer=PygmentsLexer(JavascriptLexer),
        bottom_toolbar='JavaScript edit mode. [ESC] and then [ENTER] to accept. [CTRL] + C to cancel.').strip()

    click.secho('JavaScript capture complete. Evaluating...', dim=True)

    api = state_connection.get_api()
    api.android_heap_evaluate_handle_method(target_handle, js)



import binascii
import os
import tempfile

import click
import litecli
from litecli.main import LiteCli

from ..commands.filemanager import download, upload, pwd, path_exists


def modify_config(rc):
    """
        Monkey patches the LiteCLI config to toggle
        settings that make more sense for us.

        :param rc:
        :return:
    """

    c = real_get_config(rc)
    c['main']['less_chatty'] = 'True'
    c['main']['enable_pager'] = 'False'

    return c


real_get_config = litecli.main.get_config
litecli.main.get_config = modify_config


def cleanup(p) -> None:
    """
        Remove a cached SQLite db

        :param p:
        :return:
    """

    os.remove(p)


def _should_sync_once_done(args: list) -> bool:
    """
        Checks if --sync flag was provided.

        :param args:
        :return:
    """

    return '--sync' in args


def connect(args: list) -> None:
    """
        Connects to a SQLite database by downloading a copy of the database
        from the device and storing it locally in a temporary directory.

        :param args:
        :return:
    """

    if len(args) <= 0:
        click.secho('Usage: sqlite connect <remote_file> (optional: --sync)', bold=True)
        return

    db_location = args[0]
    _, local_path = tempfile.mkstemp('.sqlite')
    use_shm = False  # does Shared Memory temp file exist ?
    use_wal = False  # does Write-Ahead-Log temp file exist ?
    use_jnl = False  # does Journal temp file exist ?
    write_back_tmp_sqlite = False  # if enabled temporary DB files are re-uploaded, this has not been testes

    # update the full remote path for future syncs
    full_remote_file = db_location \
        if os.path.isabs(db_location) else os.path.join(pwd(), db_location)

    click.secho('Caching local copy of database file...', fg='green')
    download([db_location, local_path])
    if path_exists(full_remote_file + '-shm'):
        click.secho('... caching local copy of database "shm" file...', fg='green')
        download([db_location + '-shm', local_path + '-shm'])
        use_shm = True
    if path_exists(full_remote_file + '-wal'):
        click.secho('... caching local copy of database "wal" file...', fg='green')
        download([db_location + '-wal', local_path + '-wal'])
        use_wal = True
    if path_exists(full_remote_file + '-journal'):
        click.secho('... caching local copy of database "journal" file...', fg='green')
        download([db_location + '-journal', local_path + '-journal'])
        use_jnl = True

    click.secho('Validating SQLite database format', dim=True)
    with open(local_path, 'rb') as f:
        header = f.read(16)
        header = binascii.hexlify(header)

    if header != b'53514c69746520666f726d6174203300':
        click.secho('File does not appear to be a SQLite3 db. Try downloading and manually inspecting this one.',
                    fg='red')
        cleanup(local_path)
        return

    click.secho('Connected to SQLite database at: {0}'.format(db_location), fg='green')

    # boot the litecli prompt
    lite = LiteCli(prompt='SQLite @ {} > '.format(db_location))
    lite.connect(local_path)
    lite.run_cli()

    if _should_sync_once_done(args):
        click.secho('Synchronizing changes back...', dim=True)
        upload([local_path, full_remote_file])
        # re-uploading temp sqlite files has not been tested and thus is disabled by default
        if write_back_tmp_sqlite:
            if use_shm:
                upload([local_path + '-shm', full_remote_file + '-shm'])
            if use_wal:
                upload([local_path + '-wal', full_remote_file + '-wal'])
            if use_jnl:
                upload([local_path + '-journal', full_remote_file + '-journal'])
    else:
        click.secho('NOT synchronizing changes back to device. Use --sync if you want that.', fg='green')

    # maak skoon
    cleanup(local_path)
    if use_shm:
        cleanup(local_path + '-shm')
    if use_wal:
        cleanup(local_path + '-wal')
    if use_jnl:
        cleanup(local_path + '-journal')



import click

from objection.state.connection import state_connection


def execute(args: list) -> None:
    """
        Runs a shell command on an Android device.

        :param args:
        :return:
    """

    command = ' '.join(args)
    click.secho('Running shell command: {0}\n'.format(command), dim=True)

    api = state_connection.get_api()
    response = api.android_shell_exec(command)

    if 'stdOut' in response and len(response['stdOut']) > 0:
        click.secho(response['stdOut'], bold=True)

    if 'stdErr' in response and len(response['stdErr']) > 0:
        click.secho(response['stdErr'], bold=True, fg='red')



import threading
import time
from pathlib import Path

import click
from frida import ServerNotRunningError

from objection.commands.plugin_manager import load_plugin
from objection.utils.agent import Agent, AgentConfig
from objection.utils.helpers import debug_print, warn_about_older_operating_systems
from .repl import Repl
from ..__init__ import __version__
from ..commands.mobile_packages import patch_ios_ipa, patch_android_apk, sign_android_apk
from ..state.api import api_state
from ..state.app import app_state
from ..state.connection import state_connection


def get_agent() -> Agent:
    """ get_agent bootstraps an agent instance """
    agent = Agent(AgentConfig(
        name=state_connection.name,
        host=state_connection.host,
        port=state_connection.port,
        device_type=state_connection.device_type,
        device_id=state_connection.device_id,
        spawn=state_connection.spawn,
        foremost=state_connection.foremost,
        debugger=state_connection.debugger,
        pause=not state_connection.no_pause,
        uid=state_connection.uid
    ))

    try:
        agent.run()
    except ServerNotRunningError:
        click.secho('Frida server or gadget is not running on the target!', fg='red')
        exit(1)

    return agent


# Start the Click command group
@click.group()
@click.option('--network', '-N', is_flag=True, help='Connect using a network connection instead of USB.',
              show_default=True)
@click.option('--local', '-L', is_flag=True,
              help='Connect using a local connection (for iOS Simulator).', show_default=True)
@click.option('--host', '-h', default='127.0.0.1', show_default=True)
@click.option('--port', '-P', required=False, default=27042, show_default=True)
@click.option('--api-host', '-ah', default='127.0.0.1', show_default=True)
@click.option('--api-port', '-ap', required=False, default=8888, show_default=True)
@click.option('--name', '-n', required=False,
              help='Name or bundle identifier to attach to.', show_default=True)
@click.option('--gadget', '-g', is_eager=True, hidden=True, deprecated="Please use '-n' or '--name' instead")
@click.option('--serial', '-S', required=False, default=None, help='A device serial to connect to.')
@click.option('--debug', '-d', required=False, default=False, is_flag=True,
              help='Enable debug mode with verbose output.')
@click.option('--spawn', '-s', required=False, is_flag=True, help='Spawn the target.')
@click.option('--no-pause', '-p', required=False, is_flag=True, help='Resume the target immediately.')
@click.option('--foremost', '-f', required=False, is_flag=True, help='Use the current foremost application.')
@click.option('--debugger', required=False, default=False, is_flag=True, help='Enable the Chrome debug port.')
@click.option('--uid', required=False, default=None, help='Specify the uid to run as (Android only).')
def cli(network: bool, local: bool, host: str, port: int, api_host: str, api_port: int,
        name: str, gadget: str, serial: str, debug: bool, spawn: bool, no_pause: bool,
        foremost: bool, debugger: bool, uid: int) -> None:
    """
        \b
             _   _         _   _
         ___| |_|_|___ ___| |_|_|___ ___
        | . | . | | -_|  _|  _| | . |   |
        |___|___| |___|___|_| |_|___|_|_|
              |___|(object)inject(ion)
        \b
             Runtime Mobile Exploration
                by: @leonjza from @sensepost
    """

    if debug:
        app_state.debug = debug

    if network and local:
        raise click.UsageError('The --local flag cannot be used with --network.')

    if local:
        state_connection.use_local()
        state_connection.host = None
        state_connection.port = None
    elif network:
        state_connection.use_network()
        state_connection.host = host
        state_connection.port = port
    else:
        state_connection.use_usb()
        state_connection.host = None
        state_connection.port = None

    if serial:
        state_connection.device_id = serial

    # set api parameters
    app_state.api_host = api_host
    app_state.api_port = api_port

    # Backwards compatibility
    if gadget is not None:
        name = gadget

    state_connection.name = name
    state_connection.spawn = spawn
    state_connection.no_pause = no_pause
    state_connection.foremost = foremost
    state_connection.debugger = debugger
    state_connection.uid = uid


@cli.command()
def api():
    """
        Start the objection API server in headless mode.
    """

    agent = get_agent()
    state_connection.set_agent(agent=agent)

    click.secho(f'Starting API server on {app_state.api_host}:{app_state.api_port}', fg='yellow', bold=True)
    api_state.start(app_state.api_host, app_state.api_port, app_state.debug)


@cli.command()
@click.option('--plugin-folder', '-P', required=False, default=None, help='The folder to load plugins from.')
@click.option('--quiet', '-q', required=False, default=False, is_flag=True)
@click.option('--startup-command', '-s', required=False, multiple=True,
              help='A command to run before the repl polls the device for information.')
@click.option('--file-commands', '-c', required=False, type=click.File('r'),
              help=('A file containing objection commands, separated by a '
                    'newline, that will run before the repl polls the device for information.'))
@click.option('--startup-script', '-S', required=False, type=click.File('r'),
              help='A script to import and run before the repl polls the device for information.')
@click.option('--enable-api', '-a', required=False, default=False, is_flag=True,
              help='Start the objection API server.')
def start(plugin_folder: str, quiet: bool, startup_command: str, file_commands, startup_script: click.File,
          enable_api: bool) -> None:
    """
        Start a new session
    """

    agent = get_agent()
    state_connection.set_agent(agent)

    # load plugins
    if plugin_folder:
        folder = Path(plugin_folder).resolve()
        debug_print(f'[plugin] Plugins path is: {folder}')
        for p in folder.iterdir():
            if p.is_file() or p.name.startswith('.'):
                debug_print(f'[plugin] Skipping {p.name}')
                continue

            debug_print(f'[plugin] Attempting to load plugin at {p}')
            load_plugin([p])

    repl = Repl()

    if startup_script:
        click.secho(f'Importing and running startup script at: {startup_script}', dim=True)
        script_name = f'startup_script<{startup_script.name}>'
        agent.attach_script(script_name, startup_script.read())

    if startup_command:
        for command in startup_command:
            click.secho(f'Running a startup command... {command}', dim=True)
            repl.run_command(command)

    if file_commands:
        click.secho('Running commands from file...', bold=True)
        for command in file_commands.readlines():

            command = command.strip()
            if command == '':
                continue

            # run the command using the instantiated repl
            click.secho(f'Running: \'{command}\':\n', dim=True)
            repl.run_command(command)

    warn_about_older_operating_systems()

    # start the api server
    if enable_api:
        def api_thread():
            """ Small method to run Flask non-blocking """
            api_state.start(app_state.api_host, app_state.api_port)

        click.secho(f'Starting API server on {app_state.api_host}:{app_state.api_port}', fg='yellow', bold=True)
        thread = threading.Thread(target=api_thread)
        thread.daemon = True
        thread.start()

        time.sleep(2)

    # drop into the repl
    repl.run(quiet=quiet)

# Some ugly backwards compatibility
@cli.command(deprecated="Use 'objection start' instead of 'objection explore'", hidden=True)
@click.option('--plugin-folder', '-P', required=False, default=None, help='The folder to load plugins from.')
@click.option('--quiet', '-q', required=False, default=False, is_flag=True)
@click.option('--startup-command', '-s', required=False, multiple=True,
              help='A command to run before the repl polls the device for information.')
@click.option('--file-commands', '-c', required=False, type=click.File('r'),
              help=('A file containing objection commands, separated by a '
                    'newline, that will run before the repl polls the device for information.'))
@click.option('--startup-script', '-S', required=False, type=click.File('r'),
              help='A script to import and run before the repl polls the device for information.')
@click.option('--enable-api', '-a', required=False, default=False, is_flag=True,
              help='Start the objection API server.')
def explore(plugin_folder: str, quiet: bool, startup_command: str, file_commands, startup_script: click.File,
            enable_api: bool) -> None:
    """
        Deprecated: Use 'start' instead.
    """
    # Call the start command's callback directly
    ctx = click.get_current_context()
    ctx.invoke(start,
               plugin_folder=plugin_folder,
               quiet=quiet,
               startup_command=startup_command,
               file_commands=file_commands,
               startup_script=startup_script,
               enable_api=enable_api)

@cli.command()
@click.option('--hook-debug', '-d', required=False, default=False, is_flag=True,
              help='Print compiled hooks as they are run to the screen and logfile.')
@click.argument('command', nargs=-1)
def run(hook_debug: bool, command: tuple) -> None:
    """
        Run a single objection command.
    """

    if len(command) <= 0:
        click.secho('Please specify a command to run', fg='red')
        return

    # specify if hooks should be debugged
    app_state.debug_hooks = hook_debug

    agent = get_agent()
    state_connection.set_agent(agent=agent)

    command = ' '.join(command)

    # use the methods in the main REPL to run the command
    click.secho('Running command... `{0}`'.format(command), dim=True)
    Repl().run_command(command)


@cli.command()
def version() -> None:
    """
        Prints the current version and exits.
    """

    click.secho('objection: {0}'.format(__version__))


@cli.command()
@click.option('--source', '-s', help='The source IPA to patch', required=True)
@click.option('--gadget-version', '-V', help=('The gadget version to use. If not '
                                              'specified, the latest version will be used.'), default=None)
@click.option('--codesign-signature', '-c',
              help='Codesigning Identity to use. Get it with: `security find-identity -p codesigning -v`',
              required=True)
@click.option('--provision-file', '-P', help='The .mobileprovision file to use in the patched .ipa')
@click.option('--binary-name', '-b', help='Name of the Mach-O binary in the IPA (used to patch with Frida)')
@click.option('--skip-cleanup', '-k', is_flag=True,
              help='Do not clean temporary files once finished.', show_default=True)
@click.option('--pause', '-p', is_flag=True, help='Pause the patcher before rebuilding the IPA.',
              show_default=True)
@click.option('--unzip-unicode', '-z', is_flag=True, help='Unzip IPA containing Unicode characters.')
@click.option('--gadget-config', '-C', default=None, help=(
        'The gadget configuration file to use. '
        'Refer to https://frida.re/docs/gadget/ for more information.'), show_default=False)
@click.option('--script-source', '-l', default=None, help=(
        'A script file to use with the the "path" config type. '
        'Remember that use the name of this file in your "path". It will be next to the config.'), show_default=False)
@click.option('--bundle-id', '-b', default=None, help='The bundleid to set when codesigning the IPA')
def patchipa(source: str, gadget_version: str, codesign_signature: str, provision_file: str, binary_name: str,
             skip_cleanup: bool, pause: bool, unzip_unicode: bool, gadget_config: str, script_source: str,
             bundle_id: str) -> None:
    """
        Patch an IPA with the FridaGadget dylib.
    """

    patch_ios_ipa(**locals())


@cli.command()
@click.option('--source', '-s', help='The source APK to patch', required=True)
@click.option('--architecture', '-a',
              help=('The architecture of the device the patched APK will run on. '
                    'This can be determined with `adb shell getprop ro.product.cpu.abi`. '
                    'If it is not specified, this command will try and determine it automatically.'), required=False)
@click.option('--gadget-version', '-V', help=('The gadget version to use. If not '
                                              'specified, the latest version will be used.'), default=None)
@click.option('--pause', '-p', is_flag=True, help='Pause the patcher before rebuilding the APK.',
              show_default=True)
@click.option('--skip-cleanup', '-k', is_flag=True,
              help='Do not clean temporary files once finished.', show_default=True)
@click.option('--enable-debug', '-d', is_flag=True,
              help='Set the android:debuggable flag to true in the application manifest.', show_default=True)
@click.option('--network-security-config', '-N', is_flag=True, default=False,
              help='Include a network_security_config.xml file allowing for user added CA\'s to be trusted on '
                   'Android 7 and up. This option can not be used with the --skip-resources flag.')
@click.option('--skip-resources', '-D', is_flag=True, default=False,
              help='Skip resource decoding as part of the apktool processing.', show_default=False)
@click.option('--skip-signing', '-C', is_flag=True, default=False,
              help='Skip signing the apk file.', show_default=False)
@click.option('--target-class', '-t', help='The target class to patch.', default=None)
@click.option('--use-aapt2', '-2', is_flag=True, default=False,
              help='Use the aapt2 binary instead of aapt as part of the apktool processing.', show_default=False)
@click.option('--gadget-config', '-c', default=None, help=(
        'The gadget configuration file to use. '
        'Refer to https://frida.re/docs/gadget/ for more information.'), show_default=False)
@click.option('--script-source', '-l', default=None,
              help=('A script file to use with the the "path" config type. '
                    'Specify "libfrida-gadget.script.so" as the "path" in your config.'), show_default=False)
@click.option('--ignore-nativelibs', '-n', is_flag=True, default=False,
              help='Do not change the extractNativeLibs flag in the AndroidManifest.xml.', show_default=False)
@click.option('--manifest', '-m', help='A decoded AndroidManifest.xml file to read.', default=None)
@click.option('--only-main-classes', help="Only patch classes that are in the main dex file.", is_flag=True, default=False)
@click.option('--fix-concurrency-to', '-j', help="Only use N threads for repackaging - set to 1 if running into OOM errors.", default=None)

def patchapk(source: str, architecture: str, gadget_version: str, pause: bool, skip_cleanup: bool,
             enable_debug: bool, skip_resources: bool, network_security_config: bool, target_class: str,
             use_aapt2: bool, gadget_config: str, script_source: str, ignore_nativelibs: bool, manifest: str, skip_signing: bool, only_main_classes:bool = False, fix_concurrency_to = None) -> None:
    """
        Patch an APK with the frida-gadget.so.
    """

    # ensure we decode resources if we have the network-security-config flag.
    if network_security_config and skip_resources:
        click.secho('The --network-security-config flag is incompatible with the --skip-resources flag.', fg='red')
        return

    # ensure we decode resources if we have the enable-debug flag.
    if enable_debug and skip_resources:
        click.secho('The --enable-debug flag is incompatible with the --skip-resources flag.', fg='red')
        return

    # ensure we decode resources if we do not have the --ignore-nativelibs flag.
    if not ignore_nativelibs and skip_resources:
        click.secho('The --ignore-nativelibs flag is required with the --skip-resources flag.', fg='red')
        return

    patch_android_apk(**locals())


@cli.command()
@click.argument('sources', nargs=-1, type=click.Path(exists=True), required=True)
@click.option('--skip-cleanup', '-k', is_flag=True,
              help='Do not clean temporary files once finished.', show_default=True)
def signapk(sources, skip_cleanup: bool) -> None:
    """
        Zipalign and sign an APK with the objection key.
    """
    for source in sources:
        sign_android_apk(source, skip_cleanup)


if __name__ == '__main__':
    cli()



import json
import os
import shlex
import shutil
from subprocess import list2cmdline

import click

from .github import Github

# default paths
objection_path = os.path.join(os.path.expanduser('~'), '.objection')
gadget_versions = os.path.join(objection_path, 'gadget_versions')


def list2posix_cmdline(seq):
    """
        Translate a sequence of arguments into a command line
        string.

        Implemented using shlex.quote because 
        subprocess.list2cmdline doesn't work with POSIX
    """

    return ' '.join(map(shlex.quote, seq))


class BasePlatformGadget(object):
    """ Class with base methods for any platforms Gadget downloaded """

    def __init__(self, github: Github) -> None:
        """
            Build a new instance with an existing Github instance.

            :param github:
        """

        self.github = github

    @staticmethod
    def get_local_version(gadget_type: str) -> str:
        """
            Check and return the local version of the FridaGadget
            type we have.

            :return:
        """

        if not os.path.exists(gadget_versions):
            return '0'

        with open(gadget_versions, 'r') as f:
            versions = f.read()

        # load the json.
        try:

            versions = json.loads(versions)

        except json.decoder.JSONDecodeError:
            return '0'

        if gadget_type in versions:
            return versions[gadget_type]

        return '0'

    def set_local_version(self, gadget_type: str, version: str):
        """
            Writes the version number to file, recording it as
            the current local version.

            :param gadget_type:
            :param version:
            :return:
        """

        # read the current versions if it exists, else start
        # a new dictionary for it
        if os.path.exists(gadget_versions):

            # load the json from disk
            try:

                with open(gadget_versions, 'r') as f:
                    versions = json.loads(f.read())

            except json.decoder.JSONDecodeError:
                versions = {}

        else:
            versions = {}

        # add the new version
        versions[gadget_type] = version

        # and write it to file
        with open(gadget_versions, 'w') as f:
            f.write(json.dumps(versions))

        return self


class BasePlatformPatcher(object):
    """ Base class with methods used by any platform patcher. """

    # extended classes should fill this property
    required_commands = {}

    def __init__(self):

        # check dependencies
        self.have_all_commands = self._check_commands()
        self.command_run_timeout = 60 * 5

        if os.name == 'nt':
            self.list2cmdline = list2cmdline
        else:
            self.list2cmdline = list2posix_cmdline

    def _check_commands(self) -> bool:
        """
            Check if the shell commands in required_commands are
            available.

            :return:
        """

        for cmd, attributes in self.required_commands.items():

            location = shutil.which(cmd)

            if location is None:
                click.secho('Unable to find {0}. Install it with: {1} before continuing.'.format(
                    cmd, attributes['installation']), fg='red', bold=True)

                return False

            self.required_commands[cmd]['location'] = location

        return True

    def are_requirements_met(self):
        """
            Checks if the command requirements have all been met.

            :return:
        """

        return self.have_all_commands



import os
import shutil

import click
import delegator
from packaging.version import Version

from ..utils.patchers.android import AndroidGadget, AndroidPatcher
from ..utils.patchers.github import Github
from ..utils.patchers.ios import IosGadget, IosPatcher


def patch_ios_ipa(source: str, codesign_signature: str, provision_file: str, binary_name: str,
                  skip_cleanup: bool, unzip_unicode: bool, gadget_version: str = None,
                  pause: bool = False, gadget_config: str = None, script_source: str = None,
                  bundle_id: str = None) -> None:
    """
        Patches an iOS IPA by extracting, injecting the Frida dylib,
        codesigning the dylib and app executable and rezipping the IPA.

        :param bundle_id:
        :param source:
        :param codesign_signature:
        :param provision_file:
        :param binary_name:
        :param skip_cleanup:
        :param unzip_unicode:
        :param gadget_version:
        :param pause:
        :param gadget_config:
        :param script_source:
        :return:
    """

    github = Github(gadget_version=gadget_version)
    ios_gadget = IosGadget(github)

    # get the gadget version numbers
    # check if a gadget version was specified. if not, get the latest one.
    if gadget_version is not None:
        github_version = gadget_version
        click.secho('Using manually specified version: {0}'.format(gadget_version), fg='green', bold=True)
    else:
        github_version = github.get_latest_version()
        click.secho('Using latest Github gadget version: {0}'.format(github_version), fg='green', bold=True)

    # get the local version number of the stored gadget
    local_version = ios_gadget.get_local_version('ios_universal')

    # check if the local version needs updating. this can be either because
    # the version is outdated or we simply don't have the gadget yet
    if Version(github_version) != Version(local_version) or not ios_gadget.gadget_exists():
        # download!
        click.secho('Remote FridaGadget version is v{0}, local is v{1}. Downloading...'.format(
            github_version, local_version), fg='green')

        # download, unpack, update local version and cleanup the temp files.
        ios_gadget.download() \
            .unpack() \
            .set_local_version('ios_universal', github_version) \
            .cleanup()

    click.secho('Patcher will be using Gadget version: {0}'.format(github_version), fg='green')

    # start the patching process
    patcher = IosPatcher(skip_cleanup=skip_cleanup)

    # return of we do not have all of the requirements.
    if not patcher.are_requirements_met():
        return

    patcher.set_provsioning_profile(provision_file=provision_file, bundle_id=bundle_id)
    patcher.extract_ipa(unzip_unicode, ipa_source=source)
    patcher.set_application_binary(binary=binary_name)
    patcher.patch_and_codesign_binary(
        frida_gadget=ios_gadget.get_gadget_path(), codesign_signature=codesign_signature, gadget_config=gadget_config)

    if script_source:
        click.secho('Copying over a custom script to use with the gadget config.', fg='green')
        shutil.copyfile(script_source, os.path.join(patcher.app_folder, 'Frameworks', script_source))

    # give a chance to make any last minute modifications if needed
    if pause:
        click.secho(('Patching paused. The next step is to rebuild the IPA. '
                     'If you require any manual fixes, the current temp '
                     'directory is:'), bold=True)
        click.secho('{0}'.format(patcher.app_folder), fg='green', bold=True)

        input('Press ENTER to continue...')

    patcher.archive_and_codesign(original_name=source, codesign_signature=codesign_signature)

    click.secho('Copying final ipa from {0} to current directory...'.format(patcher.get_patched_ipa_path()))
    shutil.copyfile(
        patcher.get_patched_ipa_path(),
        os.path.join(os.path.abspath('.'), os.path.basename(patcher.get_patched_ipa_path())))


def patch_android_apk(source: str, architecture: str, pause: bool, skip_cleanup: bool = True,
                      enable_debug: bool = True, gadget_version: str = None, skip_resources: bool = False,
                      network_security_config: bool = False, target_class: str = None,
                      use_aapt2: bool = False, gadget_config: str = None, script_source: str = None,
                      ignore_nativelibs: bool = True, manifest: str = None, skip_signing: bool = False,
                      only_main_classes: bool = False, fix_concurrency_to = None) -> None:
    """
        Patches an Android APK by extracting, patching SMALI, repackaging
        and signing a new APK.

        :param source:
        :param architecture:
        :param pause:
        :param skip_cleanup:
        :param enable_debug:
        :param gadget_version:
        :param skip_resources:
        :param network_security_config:
        :param target_class:
        :param use_aapt2:
        :param gadget_config:
        :param script_source:
        :param manifest:
        :param skip_signing:
        :param ignore_nativelibs:
        :param only_main_classes:
        :param fix_concurrency_to:

        :return:
    """

    github = Github(gadget_version=gadget_version)
    android_gadget = AndroidGadget(github)

    # without an architecture set, attempt to determine one using adb
    if not architecture:
        click.secho('No architecture specified. Determining it using `adb`...', dim=True)
        o = delegator.run('adb shell getprop ro.product.cpu.abi')

        # read the ach from the process' output
        architecture = o.out.strip()

        if len(architecture) <= 0:
            click.secho('Failed to determine architecture. Is the device connected and authorized?',
                        fg='red', bold=True)
            return

        click.secho('Detected target device architecture as: {0}'.format(architecture), fg='green', bold=True)

    # set the architecture we are interested in
    android_gadget.set_architecture(architecture)

    # check the gadget config flags
    if script_source and not gadget_config:
        click.secho('A script source was specified but no gadget configuration was set.', fg='red', bold=True)
        return

    # check if a gadget version was specified. if not, get the latest one.
    if gadget_version is not None:
        github_version = gadget_version
        click.secho('Using manually specified version: {0}'.format(gadget_version), fg='green', bold=True)
    else:
        github_version = github.get_latest_version()
        click.secho('Using latest Github gadget version: {0}'.format(github_version), fg='green', bold=True)

    # get local version of the stored gadget
    local_version = android_gadget.get_local_version('android_' + architecture)

    # check if the local version needs updating. this can be either because
    # the version is outdated or we simply don't have the gadget yet, or, we want
    # a very specific version
    if Version(github_version) != Version(local_version) or not android_gadget.gadget_exists():
        # download!
        click.secho('Remote FridaGadget version is v{0}, local is v{1}. Downloading...'.format(
            github_version, local_version), fg='green')

        # download, unpack, update local version and cleanup the temp files.
        android_gadget.download() \
            .unpack() \
            .set_local_version('android_' + architecture, github_version) \
            .cleanup()

    click.secho('Patcher will be using Gadget version: {0}'.format(github_version), fg='green')

    patcher = AndroidPatcher(skip_cleanup=skip_cleanup, skip_resources=skip_resources, manifest=manifest, only_main_classes=only_main_classes)

    # ensure that we have all of the commandline requirements
    if not patcher.are_requirements_met():
        return

    # ensure we have the latest apk-tool and run the
    if not patcher.is_apktool_ready():
        click.secho('apktool is not ready for use', fg='red', bold=True)
        return

    # work on patching the APK
    patcher.set_apk_source(source=source)
    patcher.unpack_apk(fix_concurrency_to=fix_concurrency_to)
    patcher.inject_internet_permission()

    if not ignore_nativelibs:
        patcher.extract_native_libs_patch()

    if enable_debug:
        patcher.flip_debug_flag_to_true()

    if network_security_config:
        patcher.add_network_security_config()

    patcher.inject_load_library(target_class=target_class)
    patcher.add_gadget_to_apk(architecture, android_gadget.get_frida_library_path(), gadget_config)

    if script_source:
        click.secho('Copying over a custom script to use with the gadget config.', fg='green')
        shutil.copyfile(script_source,
                        os.path.join(patcher.apk_temp_directory, 'lib', architecture,
                                     'libfrida-gadget.script.so'))

    # if we are required to pause, do that.
    if pause:
        click.secho(('Patching paused. The next step is to rebuild the APK. '
                     'If you require any manual fixes, the current temp '
                     'directory is:'), bold=True)
        click.secho('{0}'.format(patcher.get_temp_working_directory()), fg='green', bold=True)

        input('Press ENTER to continue...')

    patcher.build_new_apk(use_aapt2=use_aapt2, fix_concurrency_to=fix_concurrency_to)
    patcher.zipalign_apk()
    if not skip_signing:
        patcher.sign_apk()

    # woohoo, get the APK!
    destination = source.replace('.apk', '.objection.apk')

    click.secho(
        'Copying final apk from {0} to {1} in current directory...'.format(patcher.get_patched_apk_path(), destination))
    shutil.copyfile(patcher.get_patched_apk_path(), os.path.join(os.path.abspath('.'), destination))


def sign_android_apk(source: str, skip_cleanup: bool = True) -> None:
    """
        Zipaligns and signs an Android APK with the objection key.

        :param source:
        :param skip_cleanup:

        :return:
    """

    patcher = AndroidPatcher(skip_cleanup=skip_cleanup)

    # ensure that we have all of the commandline requirements
    if not patcher.are_requirements_met():
        return

    patcher.set_apk_source(source=source)
    patcher.zipalign_apk()
    patcher.sign_apk()

    # woohoo, get the APK!
    destination = source.replace('.apk', '.objection.apk')

    click.secho(
        'Copying final apk from {0} to {1} in current directory...'.format(patcher.get_patched_apk_path(), destination))
    shutil.copyfile(patcher.get_patched_apk_path(), os.path.join(os.path.abspath('.'), destination))



from objection.state.connection import state_connection


def disable(args: list = None) -> None:
    """
        Performs a generic anti root detection.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.android_root_detection_disable()


def simulate(args: list = None) -> None:
    """
        Simulate a rooted environment.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.android_root_detection_enable()



import json

import click
from tabulate import tabulate

from objection.state.connection import state_connection


def _should_output_json(args: list) -> bool:
    """
        Checks if --json is in the list of tokens received from the
        command line.

        :param args:
        :return:
    """

    return len(args) > 0 and '--json' in args


def entries(args: list = None) -> None:
    """
        Lists entries in the Android KeyStore

        :param args:
        :return:
    """

    api = state_connection.get_api()
    ks = api.android_keystore_list()

    output = [[x['alias'], x['is_key'], x['is_certificate']] for x in ks]
    click.secho(tabulate(output, headers=['Alias', 'Key', 'Certificate']))


def detail(args: list = None) -> None:
    """
        Lists details of all items in the Android KeyStore

        :param args:
        :return:
    """

    click.secho('Listing details for all items in the Android KeyStore...', dim=True)
    api = state_connection.get_api()
    ks = api.android_keystore_detail()

    if _should_output_json(args):
        click.secho(json.dumps(ks, indent=2, sort_keys=True))
        return

    output = [[
        x['keystoreAlias'],
        x['keyAlgorithm'],
        x['keySize'],
        ','.join(x['blockModes']),
        ','.join(x['encryptionPaddings']),
        ','.join(x['digests']),
        x['keyValidityStart'],
        x['origin'],
        x['purposes'],
        ','.join(x['signaturePaddings']),
        x['isInsideSecureHardware'],
    ] for x in ks]

    click.secho(tabulate(output, headers=[
        'Alias', 'Alg', 'Size', 'Modes', 'Paddings', 'Digests',
        'Validity Start', 'Origin', 'Purposes', 'Sig Paddings', 'Sec Hardware'
    ]))


def clear(args: list = None) -> None:
    """
        Clears out an Android KeyStore

        :param args:
        :return:
    """

    if not click.confirm('Are you sure you want to clear the Android keystore?'):
        return

    api = state_connection.get_api()
    api.android_keystore_clear()


def watch(args: list = None) -> None:
    """
        Watches usage of the Android KeyStore

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.android_keystore_watch()



from flask import Blueprint, jsonify, request, abort

from objection.state.connection import state_connection
from ..utils.helpers import to_snake_case

bp = Blueprint('rpc', __name__, url_prefix='/rpc')


@bp.route('/invoke/<string:method>', methods=('GET', 'POST'))
def invoke(method):
    """
        Bridge a call to the Frida RPC. Endpoints may be sourced from
        the agent's RPC exports.

        Responses are JSON encoded by default, but can be raw by adding
        ?json=false as a query string parameter.

        :param method:
        :return:
    """

    method = to_snake_case(method)

    # post requests require a little more validation, so do that
    if request.method == 'POST':

        # ensure we have some JSON formatted post data
        post_data = request.get_json(force=True, silent=True)
        if not post_data:
            return abort(jsonify(message='POST request without a valid body received'))

    try:

        rpc = state_connection.get_api()

    except Exception as e:
        return abort(jsonify(message='Failed to talk to the Frida RPC: {e}'.format(e=str(e))))

    try:

        # invoke the method based on the http request type
        if request.method == 'POST':
            response = getattr(rpc, method)(*post_data.values())

        if request.method == 'GET':
            response = getattr(rpc, method)()

        if 'json' in request.args and request.args.get('json').lower() == 'false':
            return response

    except Exception as e:
        return abort(jsonify(message='Failed to call method: {e}'.format(e=str(e))))

    return jsonify(response)



import click
from tabulate import tabulate

from objection.state.connection import state_connection
from objection.utils.helpers import pretty_concat


def _should_include_apple_bundles(args: list) -> bool:
    """
        Checks if arguments have the --include-apple-frameworks flag

        :param args:
        :return:
    """

    return len(args) > 0 and '--include-apple-frameworks' in args


def _should_print_full_path(args: list) -> bool:
    """
        Checks if arguments have the --full-path flag

        :param args:
        :return:
    """

    return len(args) > 0 and '--full-path' in args


def _is_apple_bundle(bundle: str) -> bool:
    """
        Check if a string bundle identifier is considered an Apple
        bundle based on the fact that the bundle name starts with
        the string com.apple

        :param bundle:
        :return:
    """

    # This is a bit of an assumption, but ok.
    if bundle is None:
        return False

    if bundle.startswith('com.apple'):
        return True

    return False


def show_frameworks(args: list = None) -> None:
    """
        Prints information about bundles that represent frameworks.

        https://developer.apple.com/documentation/foundation/nsbundle/1408056-allframeworks?language=objc

        :param args:
        :return:
    """

    api = state_connection.get_api()
    frameworks = api.ios_bundles_get_frameworks()

    # apply filters
    if not _should_include_apple_bundles(args):
        frameworks = [f for f in frameworks if not _is_apple_bundle(f['bundle'])]

    # Just dump it to the screen
    click.secho(tabulate(
        [[
            entry['executable'],
            entry['bundle'],
            entry['version'],
            entry['path'] if _should_print_full_path(args) else pretty_concat(entry['path'], 40, True),
        ] for entry in frameworks
        ], headers=['Executable', 'Bundle', 'Version', 'Path'],
    ))


def show_bundles(args: list = None) -> None:
    """
        Prints information about bundles that are not necessarily frameworks

        https://developer.apple.com/documentation/foundation/nsbundle/1413705-allbundles?language=objc

        :param args:
        :return:
    """

    api = state_connection.get_api()
    bundles = api.ios_bundles_get_bundles()

    # Just dump it to the screen
    click.secho(tabulate(
        [[
            entry['executable'],
            entry['bundle'],
            entry['version'],
            entry['path'] if _should_print_full_path(args) else pretty_concat(entry['path'], 40, True),
        ] for entry in bundles
        ], headers=['Executable', 'Bundle', 'Version', 'Path'],
    ))



import os

import click
from tabulate import tabulate

from objection.state.connection import state_connection
from ..utils.helpers import sizeof_fmt, clean_argument_flags


def _should_disable_exception_handler(args: list = None) -> bool:
    """
        Checks the arguments if '--no-exception-handler'
        is part of it.

        :param args:
        :return:
    """

    return len(args) > 0 and '--no-exception-handler' in args


def frida_environment(args: list = None) -> None:
    """
        Prints information about the current Frida environment.

        :param args:
        :return:
    """

    frida_env = state_connection.get_api().env_frida()

    click.secho(tabulate([
        ('Frida Version', frida_env['version']),
        ('Process Architecture', frida_env['arch']),
        ('Process Platform', frida_env['platform']),
        ('Debugger Attached', frida_env['debugger']),
        ('Script Runtime', frida_env['runtime']),
        ('Frida Heap Size', sizeof_fmt(frida_env['heap']))
    ]))


def ping(args: list = None) -> None:
    """
        Pings the agent.

        :param args:
        :return:
    """

    agent = state_connection.get_api()
    if agent.ping():
        click.secho('The agent responds ok!', fg='green')
    else:
        click.secho('The agent did not respond ok!', fg='red')


def load_background(args: list = None) -> None:
    """
        Loads a Frida script and runs it in the background.

        :param args:
        :return:
    """

    if len(clean_argument_flags(args)) <= 0:
        click.secho('Usage: import <local path to frida-script> (optional name)',
                    bold=True)
        return

    source = args[0]

    # support ~ syntax
    if source.startswith('~'):
        source = os.path.expanduser(source)

    if not os.path.isfile(source):
        click.secho('Unable to import file {0}'.format(source), fg='red')
        return

    # read the hook sources
    with open(source, 'r') as f:
        hook = ''.join(f.read())

    agent = state_connection.get_agent()
    agent.attach_script(source, hook)




import click
from tabulate import tabulate

from ..state.connection import state_connection
from ..state.device import device_state, Android, Ios


def get_environment(args: list = None) -> None:
    """
        Get information about the current environment.

        This method will call the correct runtime specific
        method to get the information that it can.

        :param args:
        :return:
    """

    if device_state.platform == Ios:
        _get_ios_environment()

    if device_state.platform == Android:
        _get_android_environment()


def _get_ios_environment() -> None:
    """
        Prints information about the iOS environment.

        This includes the current OS version as well as directories
        of interest for the current applications Documents, Library and
        main application bundle.

        :return:
    """

    paths = state_connection.get_api().env_ios_paths()

    click.secho('')
    click.secho(tabulate(paths.items(), headers=['Name', 'Path']))


def _get_android_environment() -> None:
    """
        Prints information about the Android environment.

        :return:
    """

    paths = state_connection.get_api().env_android_paths()

    click.secho('')
    click.secho(tabulate(paths.items(), headers=['Name', 'Path']))



from objection.state.connection import state_connection


def monitor(args: list = None) -> None:
    """
        Starts a new objection job that monitors the Android clipboard
        and reports on new strings found.

        :param args:
        :return:
    """

    api = state_connection.get_api()
    api.android_monitor_clipboard()
